# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from abc import abstractmethod
from bisect import bisect

from qtpy.QtWidgets import QBoxLayout
from traits.api import Callable, Int

from karabo.common.scenemodel.api import (
    BaseLayoutModel, BoxLayoutModel, FixedLayoutModel, GridLayoutChildData,
    GridLayoutModel)
from karabogui.sceneview.bases import BaseSceneAction
from karabogui.sceneview.utils import calc_bounding_rect


class CreateToolAction(BaseSceneAction):
    """An action which simplifies the creation of actions which only change
    the currently active tool in the scene view.
    """
    # A factory for the tool to be added
    tool_factory = Callable

    def perform(self, scene_view):
        """ Perform the action on a SceneView instance.
        """
        tool = self.tool_factory()
        scene_view.set_tool(tool)
        scene_view.set_cursor('cross')


class BaseLayoutAction(BaseSceneAction):
    """Base class for layout actions.

    This contains the boilerplate that all layout actions have in common.
    """

    @abstractmethod
    def create_layout(self, gui_objects, models, selection_rect):
        """Implemented by derived classes to create and add a layout.
        """

    def keyfunc(self, gui_obj):
        """A function to be passed as the key= argument of the sort() fuction.
        when sorting the GUI objects before passing their models to
        `create_layout`.

        THIS METHOD IS OPTIONAL.
        """
        raise NotImplementedError

    def perform(self, scene_view):
        selection_model = scene_view.selection_model
        # It does not make sense to create a layout for less than 2 items
        if len(selection_model) < 2:
            return

        # Get child objects from selection model
        child_objects = list(selection_model)
        try:
            child_objects.sort(key=self.keyfunc)
        except NotImplementedError:
            pass

        child_models = [obj.model for obj in child_objects]
        # Get the new layout model
        selection_rect = calc_bounding_rect(selection_model)
        layout_model = self.create_layout(child_objects, child_models,
                                          selection_rect)

        # Apply changes by removing the previous child models and add the
        # models after.
        for model in child_models:
            scene_view.remove_model(model)
        selection_model.clear_selection()

        scene_view.add_models(layout_model)
        scene_view.select_model(layout_model)


class BoxSceneAction(BaseLayoutAction):
    """A base class for actions which create a box layout
    """
    # What's the layout direction?
    direction = Int

    def create_layout(self, gui_objects, models, selection_rect):
        x, y, *_ = selection_rect
        width, height = self._calculate_size(gui_objects)
        layout_model = BoxLayoutModel(x=x, y=y, width=width, height=height,
                                      children=models,
                                      direction=self.direction)
        return layout_model

    def _calculate_size(self, gui_objects):
        # Collate widths and heights
        widths, heights = [], []
        for obj in gui_objects:
            size = obj.sizeHint()
            width, height = size.width(), size.height()
            widths.append(width)
            heights.append(height)

        # Calculate effective size by using sum and max on the widths and
        # heights depending on the direction
        get_width, get_height = sum, max
        if self.direction == QBoxLayout.TopToBottom:
            get_width, get_height = get_height, get_width

        return get_width(widths), get_height(heights)


class BoxHSceneAction(BoxSceneAction):
    """Group horizontally action
    """
    direction = QBoxLayout.LeftToRight

    def keyfunc(self, gui_obj):
        # Order by x first, then use y for tiebreaker
        geometry = gui_obj.geometry()
        return geometry.x(), geometry.y()


class BoxVSceneAction(BoxSceneAction):
    """Group vertically action
    """
    direction = QBoxLayout.TopToBottom

    def keyfunc(self, gui_obj):
        # Order by y first, then use x for tiebreaker
        geometry = gui_obj.geometry()
        return geometry.y(), geometry.x()


class GridSceneAction(BaseLayoutAction):
    """Group in grid action
    """

    def create_layout(self, gui_objects, models, selection_rect):
        x, y, width, height = selection_rect
        layout_model = GridLayoutModel(x=x, y=y, width=width, height=height)
        layout_model.children = self._collect_models(gui_objects)
        return layout_model

    @staticmethod
    def _collect_models(gui_objects):
        """Get a list of scene object models with the correct layout data
        attached.
        """

        def _reduce(positions):
            positions.sort()
            i = 1
            while i < len(positions):
                if positions[i] - positions[i - 1] > 10:
                    i += 1
                else:
                    del positions[i]
            return positions

        xs = _reduce([o.geometry().x() for o in gui_objects])
        ys = _reduce([o.geometry().y() for o in gui_objects])

        models = []
        for obj in gui_objects:
            rect = obj.geometry()
            col = bisect(xs, rect.x())
            row = bisect(ys, rect.y())
            colspan = bisect(xs, rect.right()) - col + 1
            rowspan = bisect(ys, rect.bottom()) - row + 1

            model = obj.model
            model.layout_data = GridLayoutChildData(row=row, col=col,
                                                    rowspan=rowspan,
                                                    colspan=colspan)
            models.append(model)
        return models


class GroupSceneAction(BaseLayoutAction):
    """Group without layout action
    """

    def create_layout(self, gui_objects, models, selection_rect):
        x, y, width, height = selection_rect
        model = FixedLayoutModel(x=x, y=y, width=width, height=height,
                                 children=models)
        return model


class GroupEntireSceneAction(BaseSceneAction):
    """Group entirely"""

    def perform(self, scene_view):
        """Perform entire window grouping. """


class UngroupSceneAction(BaseSceneAction):
    """Ungroup action
    """

    def perform(self, scene_view):
        ungroup(scene_view)


class SceneBringToFrontAction(BaseSceneAction):
    """Bring to front action"""

    def perform(self, scene_view):
        send_to_front(scene_view)


class SceneSendToBackAction(BaseSceneAction):
    """Send to back action"""

    def perform(self, scene_view):
        send_to_back(scene_view)


# ----------------------------------------------------------------------------
# scene_view actions.

def send_to_front(scene_view):
    """ Sends the selection_model models to the front and updates
        the scene_view """
    selection_model = scene_view.selection_model
    if len(selection_model) == 0:
        return

    for o in selection_model:
        scene_view.bring_to_front(o.model)
    scene_view.update()


def send_to_back(scene_view):
    """ Sends the selection_model models to the back and updates
        the scene_view """
    selection_model = scene_view.selection_model
    if len(selection_model) == 0:
        return
    for o in selection_model:
        scene_view.send_to_back(o.model)
    scene_view.update()


def ungroup(scene_view):
    """ Ungroups the selection_model models and add the ungrouped
        models to the view """
    selection_model = scene_view.selection_model
    unparented_models = []
    for obj in selection_model:
        model = obj.model
        if not isinstance(model, BaseLayoutModel):
            continue

        scene_view.remove_model(model)
        unparented_models.extend(model.children)

    # Clear the layout_data for all the now unparented models
    for child in unparented_models:
        child.layout_data = None

    scene_view.add_models(*unparented_models)
    selection_model.clear_selection()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.project.api import MacroModel
from karabogui.project.utils import run_macro


def test_confirmation_dialog(gui_app, mocker):
    """Check if the confirmation dialog shows up when the macro is not saved"""
    mocker.patch("karabogui.project.utils.get_macro_servers",
                 return_value=["dummy_server"])
    mocker.patch(
        "karabo.common.sanity_check.validate_macro", return_value=True)
    mocker.patch("karabogui.project.utils.get_topology")
    mocker.patch("karabogui.project.utils.get_network")

    # Macro not changed
    macro_model = MacroModel(modified=False)
    mock_question = mocker.patch(
        "karabogui.project.utils.QMessageBox.question")
    run_macro(macro_model)
    assert mock_question.call_count == 0

    # Macro changed and not saved
    macro_model.modified = True
    run_macro(macro_model)

    title = "Run Unsaved Macro?"
    text = (
        "The Macro is not saved in the Project. Do you really want to run "
        "the Macro?")
    parent = None
    assert mock_question.call_count == 1

    args = mock_question.call_args
    arg1, arg2, arg3, *_ = args[0]
    assert arg1 == parent
    assert arg2 == title
    assert arg3 == text


def test_run_macro(gui_app, mocker):
    macro_model = MacroModel()
    message_box = mocker.patch("karabogui.project.utils.messagebox")
    run_macro(macro_model)
    message = ("No (stable) Macro server found in system topology. Macro"
               " cannot be started.")
    assert message_box.show_error.call_count == 1
    message_box.show_error.assert_called_with(message)

    run_macro(macro_model)
    assert message_box.show_error.call_count == 2
    message = ("No (stable) Macro server found in system topology. "
               "Macro cannot be started.")
    message_box.show_error.assert_called_with(message)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import karabogui.icons as icons
from karabo.common.api import ServerFlags


def get_language_icon(node):
    """Return the appropriate language (`API`) icon from a `SystemTreeNode`
    """
    _SERVER_STABLE = {
        "python": icons.python,
        "bound": icons.bound,
        "cpp": icons.cpp,
        "macro": icons.macro,
    }

    _SERVER_DEVELOPMENT = {
        "python": icons.pythonDevelopment,
        "bound": icons.boundDevelopment,
        "cpp": icons.cppDevelopment,
        "macro": icons.macroDevelopment,
    }
    attrs = node.attributes
    language = attrs.get("lang", "unknown")
    develop = (attrs.get("serverFlags", 0)
               & ServerFlags.Development == ServerFlags.Development)

    icon_dict = _SERVER_DEVELOPMENT if develop else _SERVER_STABLE
    return icon_dict.get(language, None)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import argparse
import os.path as op
import time
from glob import glob

from karabo.native import Schema, decodeBinary
from karabogui.binding.api import apply_configuration, build_binding

TEST_DATA_DIR = op.join(op.dirname(__file__), 'data')
TIME_BASE = 1e6


def _iter_schemas_and_configs():
    for name in glob(op.join(TEST_DATA_DIR, '*.schema')):
        classname = op.splitext(op.basename(name))[0]
        with open(name, 'rb') as fp:
            hsh = decodeBinary(fp.read())

        config_filename = op.splitext(name)[0] + '.config'
        with open(config_filename, 'rb') as fp:
            config = decodeBinary(fp.read())

        yield Schema(name=classname, hash=hsh), config


def inner_wrapper(func, args, num_iters=10):
    """ Runs `func` `num_iters` times on the data in `read_buffer` and
    returns the avg. call duration.
    """
    start_time = time.process_time()
    for i in range(num_iters):
        func(*args)
    elapsed_time = time.process_time() - start_time

    return elapsed_time / num_iters


def benchmark_wrapper(func, args):
    total_time = 0
    perf_time, count = 0, 0
    iter_counts = (10, 100, 1000, 10000, 100000)
    for count in iter_counts:
        start_time = time.perf_counter()
        perf_time = inner_wrapper(func, args, num_iters=count)
        total_time += time.perf_counter() - start_time
        if total_time > 1.0:
            break

    return perf_time, count


def run_benchmark(version):
    API_FUNCS = {1: (build_binding, apply_configuration)}
    API_NAMES = {1: 'Current'}
    MSG = '{} took {:.2f} us/call | {:.2f} calls/s ({} iterations)'
    build_func, config_func = API_FUNCS[version]
    print('*** Running tests for', API_NAMES[version], 'API')
    for schema, config in _iter_schemas_and_configs():
        args = (schema,)
        avg_time, iters = benchmark_wrapper(build_func, args)
        print(MSG.format('Build ' + schema.name, avg_time * TIME_BASE,
                         1 / avg_time, iters))

        args = (config, build_func(schema))
        avg_time, iters = benchmark_wrapper(config_func, args)
        print(MSG.format('Config ' + schema.name, avg_time * TIME_BASE,
                         1 / avg_time, iters))


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v', '--version', default=1, type=int,
                        help='Which version to benchmark')

    args = parser.parse_args()
    if args.version in (1,):
        run_benchmark(args.version)


if __name__ == '__main__':
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QWidget

from karabo.common.api import State
from karabo.common.scenemodel.api import ColorBoolModel
from karabo.native import Bool, Configurable, Hash
from karabogui.binding.api import apply_configuration
from karabogui.indicators import STATE_COLORS
from karabogui.testing import get_class_property_proxy

from ..colorbool import DisplayColorBool


class MockQSvgWidget(QWidget):
    def load(self, svg):
        self.loaded_data = svg


class Object(Configurable):
    prop = Bool(defaultValue=True)


def test_colorbool(gui_app, mocker):
    # set up
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    model = ColorBoolModel()
    target = "karabogui.controllers.display.colorbool.SvgWidget"
    mocker.patch(target, new=MockQSvgWidget)
    proxy.value = False

    # test basics
    controller = DisplayColorBool(proxy=proxy, model=model)
    controller.create(None)
    assert controller.widget is not None

    # exercise code paths
    apply_configuration(Hash("prop", True),
                        proxy.root_proxy.binding)
    active = controller.icon.with_color(STATE_COLORS[State.ACTIVE])
    assert controller.widget.loaded_data == bytearray(active, encoding="UTF-8")

    apply_configuration(Hash("prop", False), proxy.root_proxy.binding)
    passive = controller.icon.with_color(STATE_COLORS[State.PASSIVE])
    assert controller.widget.loaded_data == bytearray(passive,
                                                      encoding="UTF-8")

    # teardown
    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.native import (
    Bool, ByteArray, Configurable, Encoding, Hash, Int8, Node, VectorInt8,
    VectorInt16)

# Reverse map of controllers.images.
TYPENUM_MAP = {
    'bool': 0,
    'char': 2,
    'int8': 4,
    'uint8': 6,
    'int16': 8,
    'uint16': 10,
    'int32': 12,
    'uint32': 14,
    'int64': 16,
    'uint64': 18,
    'float': 20,
    'float64': 22,
}


class Pixels(Configurable):
    type = Int8()
    data = ByteArray()


class Alignment(Configurable):
    offsets = VectorInt8()
    rotations = VectorInt8()


class Geometry(Configurable):
    update = Bool(False)
    pixelRegion = VectorInt8()
    alignment = Node(Alignment)


class ImageNode(Configurable):
    pixels = Node(Pixels)
    dims = VectorInt16()
    geometry = Node(Geometry)
    encoding = Int8()


class ImageNodeWithStackAxis(ImageNode):
    stackAxis = Int8()


def get_output_node(stack_axis=True):
    """Returns a node that contains an image node that could contain a
       stackAxis property"""
    image_node = ImageNodeWithStackAxis if stack_axis else ImageNode

    class OutputNodeInner(Configurable):
        image = Node(image_node, displayType='Image')

    class OutputNode(Configurable):
        data = Node(OutputNodeInner)

    return OutputNode


def get_pipeline_schema(stack_axis=True):
    """Returns a schema that contains an output node. This node has an image
       node that could contain a stackAxis property"""
    class PipelineData(Configurable):
        output = Node(get_output_node(stack_axis), displayType='OutputChannel')

    return PipelineData().getDeviceSchema()


def _get_geometry_hash(update):
    alignment_hash = Hash('offsets', [0., 0., 0.],
                          'rotations', [0., 0., 0.])
    return Hash('update', update,
                'pixelRegion', [0, 0, 1, 1],  # x0, y0, x1, y1
                'alignment', alignment_hash)


ZAXIS = 2
dimX = 40
dimY = 30


def get_image_hash(val=0, dimZ=None, *, encoding=Encoding.GRAY,
                   stack_axis=True, update=True):
    npix = dimX * dimY
    dims_val = [dimY, dimX]
    if dimZ:
        dims_val.append(dimZ)
        npix *= dimZ
    pixel_hsh = Hash('type', TYPENUM_MAP['uint8'],
                     'data', bytearray([val for _ in range(npix)]))
    img_hsh = Hash('pixels', pixel_hsh,
                   'dims', dims_val,
                   'geometry', _get_geometry_hash(update),
                   'encoding', encoding)
    if stack_axis:
        img_hsh.set('stackAxis', ZAXIS)
    return Hash('image', img_hsh)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on June 16, 2025
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from datetime import datetime
from typing import NamedTuple

import natsort
from qtpy import uic
from qtpy.QtCore import (
    QAbstractTableModel, QEvent, QModelIndex, QSortFilterProxyModel, Qt, Slot)
from qtpy.QtWidgets import (
    QAbstractItemView, QApplication, QDialog, QHeaderView, QStyle,
    QStyledItemDelegate, QStyleOptionButton)

from karabo.common.api import KARABO_PROJECT_MANAGER
from karabogui import messagebox
from karabogui.binding.api import ProxyStatus
from karabogui.indicators import get_project_device_status_icon
from karabogui.request import call_device_slot
from karabogui.singletons.api import get_config, get_topology
from karabogui.util import get_reason_parts

from .utils import get_dialog_ui

HEADER_LABELS = ["Status", "deviceId", "classId", "serverId",
                 "project_name", "Instantiate"]
STATUS_COLUMN = HEADER_LABELS.index("Status")
DEVICE_ID_COLUMN = HEADER_LABELS.index("deviceId")
CLASS_ID_COLUMN = HEADER_LABELS.index("classId")
SERVER_ID_COLUMN = HEADER_LABELS.index("serverId")
INSTANTIATE_COLUMN = HEADER_LABELS.index("Instantiate")


class ButtonDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)

    def paint(self, painter, option, index):
        button = QStyleOptionButton()
        button.rect = option.rect
        button.text = "Instantiate"
        button.state = self.button_state(index)
        QApplication.style().drawControl(QStyle.CE_PushButton, button, painter)

    def editorEvent(self, event, model, option, index):
        if event.type() == QEvent.MouseButtonRelease:
            if option.rect.contains(event.pos()):
                self.button_clicked(index)
                return True

        return False

    def button_state(self, index):
        status = index.sibling(index.row(), STATUS_COLUMN).data(
            role=Qt.UserRole)
        state = (QStyle.State_Enabled if status == ProxyStatus.OFFLINE
                 else QStyle.State_On)
        return state

    def button_clicked(self, index):
        status = index.sibling(index.row(), STATUS_COLUMN).data(
            role=Qt.UserRole)
        if not status == ProxyStatus.OFFLINE:
            return

        def device_reply(success, reply, request):
            if not success:
                reason, details = get_reason_parts(reply)
                messagebox.show_error(reason, details=details,
                                      parent=self.parent())
                return
            else:
                deviceId = request["args.deviceId"]
                widget = self.parent()
                widget.add_log(
                    f"Device <b>{deviceId}</b> instantiated successfully.")

        device_uuid = index.sibling(index.row(), INSTANTIATE_COLUMN).data(
            role=Qt.UserRole)
        deviceId = index.sibling(index.row(), DEVICE_ID_COLUMN).data(
            role=Qt.UserRole)
        class_id = index.sibling(index.row(), CLASS_ID_COLUMN).data(
            role=Qt.UserRole)
        serverId = index.sibling(index.row(), SERVER_ID_COLUMN).data(
            role=Qt.UserRole)

        call_device_slot(
            device_reply, KARABO_PROJECT_MANAGER, "slotGenericRequest",
            deviceId=deviceId,
            device_uuid=device_uuid, serverId=serverId, classId=class_id,
            type="instantiateProjectDevice")


class DeviceInstance(NamedTuple):
    status: ProxyStatus
    deviceId: str = ""
    classId: str = ""
    serverId: str = ""
    project_name: str = ""
    project_uuid: str = ""
    device_uuid: str = ""


class ProjectTopologyDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setModal(False)

        ui_file = get_dialog_ui("project_topology.ui")
        uic.loadUi(ui_file, self)

        self.delegate = ButtonDelegate(parent=self)
        self.table_model = ProjectTopologyModel()
        self.filter_model = QSortFilterProxyModel()
        self.filter_model.setFilterCaseSensitivity(False)
        self.filter_model.setFilterRole(Qt.DisplayRole)
        self.filter_model.setFilterKeyColumn(DEVICE_ID_COLUMN)
        self.filter_model.setSourceModel(self.table_model)
        self.filter_text.textChanged.connect(
            self.filter_model.setFilterFixedString)
        self.clear_button.clicked.connect(self._clear_clicked)

        table_view = self.tableView
        table_view.setModel(self.filter_model)
        table_view.setWordWrap(True)
        table_view.setAlternatingRowColors(True)
        table_view.setSortingEnabled(True)
        table_view.setSelectionBehavior(QAbstractItemView.SelectRows)
        table_view.setSelectionMode(QAbstractItemView.SingleSelection)
        table_view.sortByColumn(INSTANTIATE_COLUMN, Qt.AscendingOrder)
        table_view.setItemDelegateForColumn(INSTANTIATE_COLUMN, self.delegate)
        self.table = table_view

    def add_log(self, text: str):
        stamp = datetime.now().strftime('[%H:%M:%S]')
        item = f'{stamp} {text}'
        self.log_widget.append(item)
        bar = self.log_widget.verticalScrollBar()
        bar.setValue(bar.maximum())

    @Slot()
    def _clear_clicked(self):
        self.filter_text.setText("")

    def _convert_data(self, data):
        """Convert log data coming from the network"""
        topology = get_topology()

        def get_device(info):
            deviceId = info["device_name"]
            classId = info["device_class"]
            serverId = info["server_name"]
            project = info["project_name"]
            device_uuid = info["device_uuid"]
            server_attrs = topology.get_attributes(f"server.{serverId}")
            server_online = server_attrs is not None
            if not server_online:
                status = ProxyStatus.NOSERVER
            elif classId not in server_attrs.get("deviceClasses", []):
                status = ProxyStatus.NOPLUGIN
            else:
                attrs = topology.get_attributes(f"device.{deviceId}")
                if attrs:
                    if (attrs.get("classId") != classId
                            or attrs.get("serverId") != serverId):
                        status = ProxyStatus.INCOMPATIBLE
                    else:
                        status = ProxyStatus.ONLINE
                else:
                    status = ProxyStatus.OFFLINE

            return DeviceInstance(
                status=status, deviceId=deviceId, classId=classId,
                project_name=project, serverId=serverId,
                device_uuid=device_uuid)

        data = [get_device(info) for info in data]
        data = sorted(data, key=lambda d: d.deviceId)

        return data

    def initialize(self, data):
        domain = get_config()["domain"]
        self.info_label.setText(
            f"Showing the project topology from domain <b>{domain}</b>.")
        new = self._convert_data(data["items"])
        self.table_model.initialize(new)
        self._resize_table()
        self.table.sortByColumn(DEVICE_ID_COLUMN, Qt.AscendingOrder)

    def _resize_table(self):
        """Resize columns to contents"""
        columns = self.table_model.columnCount() - 1
        hor_header = self.table.horizontalHeader()
        hor_header.resizeSections(QHeaderView.ResizeToContents)
        hor_header.resizeSection(columns, QHeaderView.Stretch)
        ver_header = self.table.verticalHeader()
        ver_header.resizeSections(QHeaderView.ResizeToContents)


class ProjectTopologyModel(QAbstractTableModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._data = []

    def headerData(self, column, orientation, role):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return HEADER_LABELS[column]

    def getData(self):
        return self._data

    def rowCount(self, parent=QModelIndex()):
        return len(self._data)

    def columnCount(self, parent=QModelIndex()):
        return len(HEADER_LABELS)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None
        device = self._data[index.row()]
        column = index.column()
        if column == 0 and role == Qt.DecorationRole:
            return get_project_device_status_icon(device.status)
        if role == Qt.DisplayRole:
            if column > 0:
                return device[index.column()]
            else:
                return device[index.column()].name
        elif role == Qt.UserRole:
            if column == INSTANTIATE_COLUMN:
                return device.device_uuid
            return device[index.column()]
        elif role == Qt.ToolTipRole:
            status = index.sibling(index.row(), STATUS_COLUMN).data(
                role=Qt.UserRole)
            return str(status)

        return None

    def initialize(self, data):
        self.beginResetModel()
        try:
            self._data = natsort.natsorted(
                data, key=lambda device: device.deviceId)
        finally:
            self.endResetModel()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main

from karabo.native import Hash
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.panels.loggingpanel import LoggingPanel
from karabogui.singletons.mediator import Mediator
from karabogui.testing import GuiTestCase, singletons


class TestLogPanel(GuiTestCase):

    def test_log_panel(self):
        """Test the logging panel"""

        data = [Hash(
            "timestamp", "2021-07-05T15:08:35.250",
            "type", "INFO",
            "category", "Karabo_GuiServer_1",
            "message", "Hello!")]
        info = {'messages': data}

        mediator = Mediator()
        with singletons(mediator=mediator):
            panel = LoggingPanel()
            broadcast_event(KaraboEvent.LogMessages, info)
            self.process_qt_events()
            assert panel._log_widget.table_model.rowCount(None) == 1
            info = {'status': False}
            broadcast_event(KaraboEvent.NetworkConnectStatus, info)
            self.process_qt_events()
            assert panel._log_widget.table_model.rowCount(None) == 0


if __name__ == '__main__':
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from contextlib import contextmanager
from unittest import mock

from karabo.native import Hash
from karabogui.binding.api import (
    BindingRoot, DeviceProxy, ProxyStatus, apply_default_configuration,
    build_binding)
from karabogui.dialogs.configuration_preview import (
    ConfigurationFromPastPreview)
from karabogui.singletons.configuration import Configuration
from karabogui.testing import (
    get_device_schema, get_device_schema_allowed_state, singletons)


def test_dialog(gui_app):
    configuration = Hash("stringProperty", "bar",
                         "intProperty", 3,
                         )
    config_singleton = Configuration()
    with singletons(configuration=config_singleton):
        binding = build_binding(get_device_schema_allowed_state())
        proxy = DeviceProxy(binding=binding, server_id="Test",
                            status=ProxyStatus.ONLINE)
        apply_default_configuration(proxy.binding)
        dialog = ConfigurationFromPastPreview("This is the title",
                                              "The information comes here",
                                              configuration=configuration,
                                              proxy=proxy)
        assert dialog.windowTitle() == "This is the title"
        assert dialog.ui_info.text() == ("The information comes here "
                                         "The device is online.")
        assert dialog.ui_config_count.text() == (
            "Showing 2 reconfigurable parameters and 1 are currently allowed.")
        assert dialog.ui_swap.text() == "Show Changes"
        dialog._swap_view()
        assert dialog.ui_swap.text() == "Show Configuration"
        assert not dialog.ui_hide_readonly.isChecked()
        assert dialog.ui_existing.toPlainText() == (
            "\navailableScenes\n['scene']\n"
            "initOnlyString\nKarabo\n"
            "intProperty\n10\n"
            "readOnlyProperty\n0\n"
            "state\nON\n"
            "stringProperty\nfoo\n")

        assert dialog.ui_retrieved.toPlainText() == (
            "\navailableScenes\nRemoved from configuration\n"
            "initOnlyString\nRemoved from configuration\n"
            "intProperty\n3\n"
            "readOnlyProperty\nRemoved from configuration\n"
            "state\nRemoved from configuration\n"
            "stringProperty\nbar\n")
        assert dialog.ui_text_info_all.toPlainText() == (
            "\nstringProperty\nbar\n"
            "intProperty\n3\n"
        )
        assert dialog.ui_text_info_configurable.toPlainText() == (
            "\nstringProperty\nbar\n"
            "intProperty\n3\n"
        )

        save_path = "karabogui.dialogs.configuration_preview." \
                    "getSaveFileName"
        open_path = "karabogui.dialogs.configuration_preview.open"
        write_xml = "karabogui.dialogs.configuration_preview.writeXML"

        @contextmanager
        def open_file(fn, mode):
            """A context manager to make sure we do not write something
            """
            yield fn

        with mock.patch(save_path) as save_dialog, \
                mock.patch(open_path, new=open_file):
            with mock.patch(write_xml) as writer:
                save_dialog.return_value = "ConfigPath"
                dialog.save_configuration_to_file()
                # Check first call
                args = writer.call_args[0]
                # First argument is a Hash with unknown class here
                h = args[0]["unknown-class"]
                assert h["stringProperty"] == "bar"


def test_dialog_empty_configuration(gui_app):
    """Test the conflict with an empty configuration in saving"""
    configuration = Hash()
    config_singleton = Configuration()
    with singletons(configuration=config_singleton):
        binding = build_binding(get_device_schema())
        proxy = DeviceProxy(binding=binding, server_id="Test",
                            status=ProxyStatus.ONLINE)
        apply_default_configuration(proxy.binding)
        dialog = ConfigurationFromPastPreview("This is the title",
                                              "The information comes here",
                                              configuration=configuration,
                                              proxy=proxy)

        save_path = "karabogui.dialogs.configuration_preview." \
                    "getSaveFileName"
        mbox = "karabogui.dialogs.configuration_preview.messagebox"
        with mock.patch(save_path) as save_dialog:
            with mock.patch(mbox) as mb:
                save_dialog.return_value = "ConfigPath"
                dialog.save_configuration_to_file()
                mb.show_error.assert_called()


def test_dialog_empty_binding(gui_app):
    """Test the empty binding of the dialog protetion"""
    configuration = Hash()
    config_singleton = Configuration()
    with singletons(configuration=config_singleton):
        binding = BindingRoot()
        binding.class_id = "unknown-class"
        proxy = DeviceProxy(binding=binding, server_id="Test",
                            status=ProxyStatus.ONLINE)
        dialog = ConfigurationFromPastPreview("This is the title",
                                              "The information comes here",
                                              configuration=configuration,
                                              proxy=proxy)
        assert dialog.ui_existing.toPlainText() == (
            "No schema available to extract a configuration")
        assert dialog.ui_retrieved.toPlainText() == (
            "No schema available for comparing configurations!")


def test_dialog_only_reconfigurable_changes(gui_app):
    configuration = Hash("stringProperty", "bar",
                         "intProperty", 3,
                         "readOnlyProperty", 4,
                         )
    config_singleton = Configuration()
    with singletons(configuration=config_singleton):
        binding = build_binding(get_device_schema_allowed_state())
        proxy = DeviceProxy(binding=binding, server_id="Test",
                            status=ProxyStatus.ONLINE)
        apply_default_configuration(proxy.binding)
        dialog = ConfigurationFromPastPreview(
            "This is the title",
            "The information comes here",
            configuration=configuration,
            proxy=proxy)

        # Show all changes
        dialog._show_configuration_changes(hide_readonly=False)
        assert dialog.ui_existing.toPlainText() == (
            "\navailableScenes\n['scene']\n"
            "initOnlyString\nKarabo\n"
            "intProperty\n10\n"
            "readOnlyProperty\n0\n"
            "state\nON\n"
            "stringProperty\nfoo\n")
        assert dialog.ui_retrieved.toPlainText() == (
            "\navailableScenes\nRemoved from configuration\n"
            "initOnlyString\nRemoved from configuration\n"
            "intProperty\n3\n"
            "readOnlyProperty\n4\n"
            "state\nRemoved from configuration\n"
            "stringProperty\nbar\n")

        # Show only reconfigurable changes.
        dialog._show_configuration_changes(hide_readonly=True)
        assert dialog.ui_existing.toPlainText() == (
            "\navailableScenes\n['scene']\n"
            "initOnlyString\nKarabo\n"
            "intProperty\n10\n"
            "state\nON\n"
            "stringProperty\nfoo\n")

        assert dialog.ui_retrieved.toPlainText() == (
            "\navailableScenes\nRemoved from configuration\n"
            "initOnlyString\nRemoved from configuration\n"
            "intProperty\n3\n"
            "state\nRemoved from configuration\n"
            "stringProperty\nbar\n")



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on August 3, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from collections import OrderedDict
from contextlib import contextmanager
from functools import partial

from qtpy.QtCore import (
    QItemSelection, QItemSelectionModel, QModelIndex, QPoint, QRect, Qt,
    Signal, Slot)
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import (
    QAbstractItemDelegate, QAbstractItemView, QAction, QMenu, QTreeView)

from karabo.common.api import (
    KARABO_SCHEMA_ACCESS_MODE, KARABO_SCHEMA_ALIAS,
    KARABO_SCHEMA_ARCHIVE_POLICY, KARABO_SCHEMA_ASSIGNMENT,
    KARABO_SCHEMA_DEFAULT_VALUE, KARABO_SCHEMA_DESCRIPTION,
    KARABO_SCHEMA_MAX_EXC, KARABO_SCHEMA_MAX_INC, KARABO_SCHEMA_MAX_SIZE,
    KARABO_SCHEMA_METRIC_PREFIX_SYMBOL, KARABO_SCHEMA_MIN_EXC,
    KARABO_SCHEMA_MIN_INC, KARABO_SCHEMA_MIN_SIZE,
    KARABO_SCHEMA_REQUIRED_ACCESS_LEVEL, KARABO_SCHEMA_TAGS,
    KARABO_SCHEMA_UNIT_SYMBOL)
from karabo.native import AccessLevel, AccessMode, Assignment
from karabogui import icons
from karabogui.binding.api import (
    BaseBinding, DeviceProxy, PropertyProxy, VectorHashBinding)
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.generic_scenes import get_property_proxy_model
from karabogui.widgets.popup import PopupWidget

from .edit_delegate import EditDelegate
from .filter_model import ConfiguratorFilterModel
from .qt_item_model import ConfigurationTreeModel
from .slot_delegate import SlotButtonDelegate
from .utils import get_proxy_value
from .value_delegate import ValueDelegate


class ConfigurationTreeView(QTreeView):
    """A tree view for `Configuration` instances"""
    itemSelectionChanged = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setDragEnabled(True)
        self.setTabKeyNavigation(True)
        self.header().setSectionsMovable(False)
        self.header().setMinimumSectionSize(100)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        triggers = (QAbstractItemView.CurrentChanged |
                    QAbstractItemView.SelectedClicked)
        self.setEditTriggers(triggers)

        self.edit_delegate = EditDelegate(parent=self)

        model = ConfigurationTreeModel(parent=self)
        self._source_model = model
        self.setModel(model)
        model.dataChanged.connect(self._update_popup_contents)

        # On demand we can have a filter model
        self._filter_model = None

        # Add a delegate for rows with slot buttons
        self.slot_delegate = SlotButtonDelegate(parent=self)
        self.setItemDelegateForColumn(0, self.slot_delegate)
        # ... and a delegate for the value column
        self.setItemDelegateForColumn(1, ValueDelegate(parent=self))
        # ... and a delegate for the editable value column
        self.setItemDelegateForColumn(2, self.edit_delegate)

        # Widget for more information of an index
        self.popup_widget = None
        self._popup_showing_index = None

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_context_menu)
        self.setUniformRowHeights(True)

    # ------------------------------------
    # Public methods

    def swap_models(self):
        """Swap the models of the tree view between filter and source"""
        self.close_popup_widget()
        current_index = self.currentIndex()
        if self._filter_model is None:
            model = ConfiguratorFilterModel(self.sourceModel())
            self.setModel(model)
            self._filter_model = model
            index = model.mapFromSource(current_index)
        else:
            index = self.model().mapToSource(current_index)
            self.setModel(self.sourceModel())
            self._filter_model.deleteLater()
            self._filter_model = None
        # Select row of index and set current index
        self.setRowSelection(index)

    @contextmanager
    def assign_property_proxy(self):
        """Context manager to maintain a proxy selection"""
        model = self.model()
        try:
            index = self.currentIndex()
            proxy = model.index_ref(index)
            yield
        finally:
            if proxy is None:
                # XXX: Avoid a full model search for `None`
                return
            root_index = model.index(0, 0, QModelIndex())
            matches = model.match(root_index, Qt.UserRole, proxy.path, 1,
                                  Qt.MatchRecursive)
            index = matches[0] if matches else QModelIndex()
            self.setCurrentIndex(index)

    def sourceModel(self):
        """Public method to retrieve the source model"""
        return self._source_model

    def filterModel(self):
        """Public method to retrieve the filter model"""
        return self._filter_model

    def setRowSelection(self, index):
        """Select the full row in the configurator and set the `index`"""
        if not index.isValid():
            self.selectionModel().clearSelection()
            return

        model = index.model()
        parent = index.parent()
        selection = QItemSelection(model.index(index.row(), 0, parent),
                                   model.index(index.row(), 2, parent))
        self.selectionModel().select(selection,
                                     QItemSelectionModel.ClearAndSelect)
        self.setCurrentIndex(index)

    @contextmanager
    def assign_popup(self):
        """Assign the existing popup index to a new one"""
        try:
            index = self._popup_showing_index
            if index is not None:
                parent = index.parent()
                if not parent.isValid():
                    index = index.row(), index.column(), parent
                    self._popup_showing_index = QModelIndex()
                else:
                    index = None
                    self.close_popup_widget()
            yield
        finally:
            if index is not None:
                index = self.model().index(*index)
                if index.isValid():
                    self._popup_showing_index = index
                else:
                    self.close_popup_widget()

    def assign_proxy(self, proxy):
        with self.assign_popup():
            model = self.sourceModel()
            model.root = proxy
            if proxy is None:
                return

            # Show second column only for devices
            if isinstance(proxy, DeviceProxy):
                self.setColumnHidden(1, False)
            else:
                self.setColumnHidden(1, True)
            # XXX: This is slightly hacky, but it keeps the buttons consistent
            model.notify_of_modifications()

    def apply_all(self):
        self.sourceModel().apply_changes()

    def decline_all(self):
        model = self.sourceModel()
        model.decline_changes()
        model.announceDataChanged()

    def clear(self):
        self.assign_proxy(None)
        self.close_popup_widget()

    def close_popup_widget(self):
        if self.popup_widget is not None:
            self.popup_widget.close()
            self.popup_widget = None
        self._popup_showing_index = None

    # ------------------------------------
    # Private methods

    def _get_popup_info(self, index):
        # Get the index's stored object
        obj = self.model().index_ref(index)
        if obj is None:
            return

        # Attributes show the info of their parent proxy
        if isinstance(obj, BaseBinding):
            index = index.parent()
            obj = self.model().index_ref(index)
            if obj is None:
                return

        binding = obj.binding
        attributes = binding.attributes
        property_name = index.data(Qt.DisplayRole)
        info = OrderedDict()
        if property_name:
            info['Property'] = property_name
        if KARABO_SCHEMA_DESCRIPTION in attributes:
            info['Description'] = attributes.get(KARABO_SCHEMA_DESCRIPTION)

        info['Key'] = obj.path
        if isinstance(binding, BaseBinding):
            # XXX: Maybe make this a little different looking
            name = type(binding).__name__[:-len('binding')]
            info['Value Type'] = name
        if KARABO_SCHEMA_DEFAULT_VALUE in attributes:
            info['Default Value'] = attributes.get(KARABO_SCHEMA_DEFAULT_VALUE)
        if KARABO_SCHEMA_ACCESS_MODE in attributes:
            info['AccessMode'] = AccessMode(
                attributes[KARABO_SCHEMA_ACCESS_MODE]).name
        if KARABO_SCHEMA_REQUIRED_ACCESS_LEVEL in attributes:
            info['AccessLevel'] = AccessLevel(attributes.get(
                KARABO_SCHEMA_REQUIRED_ACCESS_LEVEL)).name
        if KARABO_SCHEMA_ASSIGNMENT in attributes:
            info['Assignment'] = Assignment(
                attributes[KARABO_SCHEMA_ASSIGNMENT]).name
        if KARABO_SCHEMA_ALIAS in attributes:
            info['Alias'] = attributes.get(KARABO_SCHEMA_ALIAS)
        if KARABO_SCHEMA_TAGS in attributes:
            info['Tags'] = ", ".join(attributes.get(KARABO_SCHEMA_TAGS))
        if binding.timestamp is not None:
            info['Timestamp'] = binding.timestamp.toLocal()
            info['Time Id'] = binding.timestamp.tid
        displayType = binding.displayType
        if displayType:
            if displayType.startswith('bin|'):
                info['Bits'] = displayType[4:]
            else:
                info["displayType"] = displayType
        if isinstance(obj.root_proxy, DeviceProxy):
            info['Value on device'] = get_proxy_value(index, obj)

        # Unit related attributes
        additional_attrs = (KARABO_SCHEMA_METRIC_PREFIX_SYMBOL,
                            KARABO_SCHEMA_UNIT_SYMBOL)
        for attr_name in additional_attrs:
            attr = attributes.get(attr_name)
            info[attr_name] = attr

        # Other additional attributes
        additional_attrs = [
            (KARABO_SCHEMA_MIN_EXC, KARABO_SCHEMA_MIN_EXC),
            (KARABO_SCHEMA_MAX_EXC, KARABO_SCHEMA_MAX_EXC),
            (KARABO_SCHEMA_MIN_INC, KARABO_SCHEMA_MIN_INC),
            (KARABO_SCHEMA_MAX_INC, KARABO_SCHEMA_MAX_INC),
            (KARABO_SCHEMA_MIN_SIZE, KARABO_SCHEMA_MIN_SIZE),
            (KARABO_SCHEMA_MAX_SIZE, KARABO_SCHEMA_MAX_SIZE),
            ('ArchivePolicy', KARABO_SCHEMA_ARCHIVE_POLICY)]

        for label, attr_name in additional_attrs:
            attr = attributes.get(attr_name)
            if attr is None:
                continue
            info[label] = attr

        return info

    def _show_popup_widget(self, index, event_pos):
        # Only if the icon was clicked
        # Get the tree widget's x position
        tree_x = self.header().sectionViewportPosition(0)
        # Get the x coordinate of the root index in order to calculate
        # the identation of the index
        root_x = self.visualRect(self.rootIndex()).x()
        # Get the rectangle of the viewport occupied by index
        vis_rect = self.visualRect(index)
        # Calculate the x coordinate of the item
        index_x = tree_x + vis_rect.x() - root_x
        # Get the rect surrounding the icon
        icon_rect = QRect(index_x, vis_rect.y(), vis_rect.height(),
                          vis_rect.height())

        if not icon_rect.contains(event_pos):
            return

        if self.popup_widget is None:
            self.popup_widget = PopupWidget(can_freeze=True, parent=self)
        else:
            # NOTE: Reset pop-up widget, e.g. clean cache
            self.popup_widget.reset()

        info = self._get_popup_info(index)
        self._popup_showing_index = index
        self.popup_widget.setInfo(info)

        pos = QCursor.pos()
        pos.setX(pos.x() + 10)
        pos.setY(pos.y() + 10)
        self.popup_widget.move(pos)
        self.popup_widget.show()

    def _selected_editable_index(self):
        """Return the currently selected editable index
        """
        indexes = self.selectionModel().selectedIndexes()
        if indexes:
            last_column = self.model().columnCount() - 1
            index = [idx for idx in indexes if idx.column() == last_column][0]
            if index.flags() & Qt.ItemIsEditable == Qt.ItemIsEditable:
                return index

    @Slot(QPoint)
    def _show_context_menu(self, event):
        """Show a custom context menu
        """
        selected_index = self._selected_editable_index()
        if selected_index is None:
            return None

        selected_obj = self.model().index_ref(selected_index)
        binding = getattr(selected_obj, 'binding', None)
        if (isinstance(selected_obj, PropertyProxy) and binding is not None and
                KARABO_SCHEMA_DEFAULT_VALUE in binding.attributes):
            menu = QMenu()
            text = "Reset to default"
            acResetToDefault = QAction(icons.revert, text, None)
            acResetToDefault.setStatusTip(text)
            acResetToDefault.setToolTip(text)
            acResetToDefault.setIconVisibleInMenu(True)
            acResetToDefault.triggered.connect(partial(self._reset_to_default,
                                                       selected_obj))
            menu.addAction(acResetToDefault)
            menu.exec(QCursor.pos())

    @Slot(object)
    def _reset_to_default(self, proxy):
        """Reset the value of the given `proxy` to the default
        """
        binding = proxy.binding
        if binding is not None:
            default_value = binding.attributes.get(KARABO_SCHEMA_DEFAULT_VALUE)
            if isinstance(proxy.root_proxy, DeviceProxy):
                proxy.edit_value = default_value
            else:
                proxy.value = default_value
            self.sourceModel().announceDataChanged()

    @Slot(QModelIndex, QModelIndex)
    def _update_popup_contents(self, topLeft, bottomRight):
        """When the data in the model changes, blindly update the popup widget
        if it happens to be showing.
        """
        if self.popup_widget is None or not self.popup_widget.isVisible():
            return

        if self._popup_showing_index.isValid():
            info = self._get_popup_info(self._popup_showing_index)
            if info is None:
                # index wasn't actually valid because the device proxy changed
                self.popup_widget.hide()
            else:
                self.popup_widget.setInfo(info)

    # ------------------------------------
    # Event handlers

    def closeEditor(self, editor, hint):
        """XXX: PyQt does not send this signal properly in the item delegate,
        so we avoid signals altogether and call it directly...
        """
        self.edit_delegate.close_editor(editor, hint)
        super().closeEditor(editor, hint)

    def currentChanged(self, current, previous):
        """Pass selection changes along to the value delegate
        """
        self.edit_delegate.current_changed(current)
        super().currentChanged(current, previous)

    def setAccessLevel(self, level: AccessLevel):
        """Set the Global Access Level"""
        model = self.sourceModel()
        proxy = model.root
        self.assign_proxy(None)
        model.setAccessLevel(level)
        self.slot_delegate.setAccessLevel(level)
        self.assign_proxy(proxy)

    def setMode(self, expert: bool):
        with self.assign_property_proxy():
            model = self.sourceModel()
            proxy = model.root
            self.assign_proxy(None)
            model.setMode(expert)
            self.assign_proxy(proxy)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            index = self.indexAt(event.pos())
            if index.isValid() and index.column() == 0:
                self._show_popup_widget(index, event.pos())
        super().mousePressEvent(event)

    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton:
            index = self.indexAt(event.pos())
            if index.isValid() and index.column() == 1:
                proxy = self.model().index_ref(index)
                model = get_property_proxy_model(proxy, include_images=True)
                if model is not None:
                    data = {"model": model}
                    broadcast_event(KaraboEvent.ShowUnattachedController, data)
                event.accept()

        super().mouseDoubleClickEvent(event)

    def keyPressEvent(self, event):
        """Reimplemented function of QTreeView.

        Emulates the QTreeView enter/escape key behavior for rows containing
        tables. NOTE: All other row types are ignored!
        """
        key_event = event.key()
        if key_event in (Qt.Key_Return, Qt.Key_Enter, Qt.Key_Escape):
            # Get current index
            selected_index = self._selected_editable_index()
            if selected_index is not None:
                # NOTE: Ensure this is a VectorHash!
                obj = self.model().index_ref(selected_index)
                is_table = isinstance(getattr(obj, 'binding', None),
                                      VectorHashBinding)
                # Act on that item
                if is_table and key_event in (Qt.Key_Return, Qt.Key_Enter):
                    self.edit_delegate.update_model_data(
                        selected_index, QAbstractItemDelegate.SubmitModelCache)
                    return
                elif key_event == Qt.Key_Escape:
                    self.edit_delegate.update_model_data(
                        selected_index, QAbstractItemDelegate.RevertModelCache)
                    return

        super().keyPressEvent(event)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import argparse
import re
import sys

from qtpy.QtCore import QObject, QTimer, Slot

import karabogui.const as global_constants
from karabo.common.api import Capabilities
from karabogui import messagebox
from karabogui.programs.base import create_gui_app, init_gui
from karabogui.programs.utils import close_app
from karabogui.request import get_scene_from_server
from karabogui.singletons.api import get_network, get_topology

DEVSCENE_PROG = re.compile(r"([0-9a-zA-Z/_\-]+)\|(.+)")
CAPA = Capabilities.PROVIDES_SCENES


class DeviceWaiter(QObject):
    """
    An helper class that listens to the system topology and acts accordingly.

    The constructor will parse the list of sceneIds for the theather as well
    """

    def __init__(self, scene_ids, timeout, parent=None):
        super().__init__(parent=parent)
        # a dictionary of list of scene names indexed by deviceId
        self.device_scenes = {}
        # track if scenes are open
        self.no_scenes = True
        self.not_capable_devices = []
        self.parse_scene_ids(scene_ids)
        self.timeout = timeout

        # Attach event to topology
        topology = get_topology()
        topology.system_tree.on_trait_change(
            self._topo_update, 'initialized')

    def is_valid(self):
        return len(self.device_scenes) > 0

    def parse_scene_ids(self, scene_ids):
        """parses the sceneIds and will provide a messagebox for errors"""
        for scene_id in scene_ids:
            match = DEVSCENE_PROG.match(scene_id)
            if match and all(match.groups()):
                device_id = match.group(1)
                scene_name = match.group(2)
                scene_names = self.device_scenes.setdefault(device_id, set())
                scene_names.add(scene_name)
        # no scenes are matching the regex, exit!.
        if not self.device_scenes:
            msg = "{} are not valid sceneIds.".format(
                ','.join([scene for scene in scene_ids]))
            messagebox.show_warning(msg, title='Theater')
            close_app()

    def start(self):
        self.timer = QTimer(self)
        # timeout in case devices are not found.
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self._timeout_handler)
        self.timer.start(self.timeout * 1000)

    def open_scenes(self, device_id, scene_names):
        for scene_name in scene_names:
            get_scene_from_server(device_id, scene_name)
            self.no_scenes = False

    @Slot()
    def _timeout_handler(self):
        # stop listening to the topology
        topology = get_topology()
        topology.system_tree.on_trait_change(
            self._topo_update, 'initialized', remove=True)

        # this dialog building is rather complex but will avoid
        # multiple error dialogs to the user.
        msg = ""
        if self.device_scenes:
            pending_scenes = []
            for device_id, scene_names in self.device_scenes.items():
                pending_scenes.extend([f'{device_id}|{scene}'
                                       for scene in scene_names])
            msg += "Could not open the following scenes: {}. \n".format(
                ',\n'.join(pending_scenes))

        missing = ',\n'.join([device_id for device_id in self.device_scenes])
        if missing:
            msg += "The following deviceIds are not present " \
                   "in the system topology: {}. \n".format(missing)
        incapables = ',\n'.join([device_id
                                 for device_id in self.not_capable_devices])
        if incapables:
            msg += "The following deviceIds do not provide " \
                   "scenes: {}. \n".format(incapables)

        # open a dialog for the missing scenes
        if msg:
            msg += "Please review command line arguments accordingly"
            messagebox.show_warning(msg, title='Theater')
        # no scenes are open after the timeout.
        if self.no_scenes:
            close_app()

    def _topo_update(self):
        topology = get_topology()
        topology.system_tree.on_trait_change(
            self._topo_update, 'initialized', remove=True)

        def alive_visitor(node):
            if node.node_id not in self.device_scenes:
                return
            scenes = self.device_scenes.pop(node.node_id)
            if (node.capabilities & CAPA) == CAPA:
                self.open_scenes(node.node_id, scenes)
            else:
                self.not_capable_devices.append(node.node_id)

        topology.visit_system_tree(alive_visitor)
        # Warning message if provided device is invalid or has no scene.
        non_exisiting_device = False
        no_capable_service = False
        msg = ""
        if self.no_scenes:
            missing = ',\n'.join(
                [device_id for device_id in self.device_scenes])
            if missing:
                msg += "The following deviceIds are not present " \
                       f"in the system topology: '{missing}'. \n\n"
                non_exisiting_device = True
        if self.not_capable_devices:
            incapables = ',\n'.join([device_id for device_id in
                                     self.not_capable_devices])
            if incapables:
                msg += ("The following deviceIds do not provide scenes: "
                        f"'{incapables}'. \n\n")
            no_capable_service = True
        if msg:
            msg += "Please review command line arguments accordingly."
            messagebox.show_warning(msg, title='Theater')
        #  All the specified devices are not available in the topology or have
        #  no scene,  exit!
        if non_exisiting_device and no_capable_service:
            close_app()


def create_theatre(ns):
    app = create_gui_app(sys.argv)
    init_gui(app, use_splash=not ns.nosplash)

    global_constants.APPLICATION_MODE = True
    network = get_network()
    waiter = DeviceWaiter(ns.scene_id, ns.timeout)

    @Slot()
    def _connect_handler(status):
        """Connect handler for direct connect attempts without dialog"""
        network.signalServerConnectionChanged.disconnect(_connect_handler)
        if not status:
            messagebox.show_error("Error, could not connect to gui server "
                                  f"<b>{ns.host}:{ns.port}</b>. "
                                  "Closing karabo theare.")
            close_app()
        else:
            # We are connected and charge a timeout now!
            waiter.start()

    # We might want to connect directly to the gui server
    if ns.host and ns.port:
        network.signalServerConnectionChanged.connect(_connect_handler)
        success = get_network().connectToServerDirectly(
            hostname=ns.host, port=ns.port)
    else:
        # Connect to the GUI Server via dialog
        success = network.connectToServer()

    return success, waiter, app


def run_theatre(ns):
    """
    The theatre downloads and opens scenes provided by devices.

    All scenes have the name deviceId|sceneName and are not editable!
    """
    success, waiter, app = create_theatre(ns)
    if success and waiter.is_valid():
        app.exec()
        app.deleteLater()
        sys.exit()
    else:
        close_app()


def main():
    ap = argparse.ArgumentParser(description='Karabo Theater')
    ap.add_argument('scene_id', type=str, nargs='+',
                    help='The scene ids. Device provided scene '
                         'formatted like: "deviceId|sceneName".')
    ap.add_argument('-host', '--host', type=str,
                    help='The hostname of the gui server to connect')
    ap.add_argument('-port', '--port', type=int,
                    help='The port number of the gui server to connect')
    ap.add_argument('-timeout', '--timeout', type=int, default=10,
                    help='The timeout in seconds to be waited for the devices'
                         'to be visible')
    ap.add_argument('-nosplash', '--nosplash', action='store_true')
    run_theatre(ap.parse_args())


if __name__ == '__main__':
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest.mock import Mock

from traits.api import pop_exception_handler, push_exception_handler

from karabo.common.api import InstanceStatus
from karabo.native import Configurable, Hash, String
from karabogui.binding.api import ProxyStatus
from karabogui.testing import (
    GuiTestCase, get_device_schema, singletons, system_hash)
from karabogui.topology.system_topology import SystemTopology
from karabogui.topology.util import (
    get_macro_servers, getTopology, is_device_online, is_server_online)


class FooClass(Configurable):
    val = String()


def setUp():
    push_exception_handler(lambda *args: None, reraise_exceptions=True)


def tearDown():
    pop_exception_handler()


class TestSystemTopology(GuiTestCase):

    def test_cleanup_topology(self):
        topology = SystemTopology()
        topology.initialize(system_hash())
        assert topology.online
        topo = getTopology()
        assert isinstance(topo, Hash)
        topology.clear()
        assert not topology.online
        assert isinstance(topo, Hash)

    def test_get_class_simple(self):
        network = Mock()
        topology = SystemTopology()
        topology.initialize(system_hash())
        with singletons(network=network, topology=topology):
            klass = topology.get_class("swerver", "NoClass")
            # server with name "swerver" don"t have the requested class
            assert klass.status is ProxyStatus.NOPLUGIN
            # GUI"s system topology won"t request for schema if the devic
            # class is not in the deviceClasses list of the server, this
            # solves the race condition in MDL
            assert network.onGetClassSchema.call_count == 0

            # Test a class that is available, request multiple times, but
            # only a single schema is requested
            foo_klass = topology.get_class("swerver", "FooClass")
            assert foo_klass.status == ProxyStatus.REQUESTED
            assert network.onGetClassSchema.call_count == 1
            more_foo_klass = topology.get_class("swerver", "FooClass")
            assert more_foo_klass.status == ProxyStatus.REQUESTED
            assert network.onGetClassSchema.call_count == 1

            # Make sure we can request when server leaves topology
            assert "FooClass" in topology._requested_classes["swerver"]
            topology.instance_gone(Hash("server", Hash("swerver", "")))
            assert "swerver" not in topology._requested_classes

    def test_get_device_simple(self):
        network = Mock()
        topology = SystemTopology()
        topology.initialize(system_hash())
        with singletons(network=network, topology=topology):
            klass = topology.get_class("swerver", "divvy")

            assert klass.status is ProxyStatus.NOPLUGIN
            assert network.onGetClassSchema.call_count == 0

            dev = topology.get_device("divvy")

            assert dev.status is ProxyStatus.ONLINEREQUESTED
            network.onGetDeviceSchema.assert_called_with("divvy")
            assert not len(dev.binding.value)
            topology.device_schema_updated("divvy", get_device_schema())
            assert len(dev.binding.value)

    def test_get_project_device_simple(self):
        network = Mock()
        topology = SystemTopology()
        topology.initialize(system_hash())
        with singletons(network=network, topology=topology):
            device_id = "divvy"
            server_id = "swerver"
            class_id = "FooClass"

            device = topology.get_project_device(device_id,
                                                 server_id=server_id,
                                                 class_id=class_id, )
            # Mocked network is not providing online proxy its schema
            device._online_proxy.status = ProxyStatus.OFFLINE
            assert device._online_proxy is topology.get_device(device_id)
            project_proxy = topology.get_project_device_proxy(
                device_id, server_id, class_id)
            assert device._offline_proxy is project_proxy
            assert device._offline_proxy.device_id == device_id
            assert device._offline_proxy.binding.class_id == class_id
            # No schema is requested for the online device
            assert network.onGetDeviceSchema.call_count == 0

            device.rename(device_id="davey")
            key = (server_id, class_id)
            assert device_id not in topology._project_device_proxies[key]
            assert "davey" in topology._project_device_proxies[key]

            network.reset_mock()
            device.rename(device_id="junk", server_id="notthere",
                          class_id="NotValieEither")
            assert network.onGetDeviceSchema.call_count == 0

            assert len(topology._project_device_proxies) == 1
            topology.remove_project_device_proxy(device_id="junk",
                                                 server_id="notthere",
                                                 class_id="NotValieEither")
            assert len(topology._project_device_proxies) == 0

    def test_system_topology_gone(self):
        network = Mock()
        topology = SystemTopology()
        topology.initialize(system_hash())
        with singletons(network=network, topology=topology):
            # 1. Plainly test a faulty changes Hash
            assert topology.get_attributes("device.divvy") is not None
            changes = Hash("new", Hash(), "update", Hash(), "gone", Hash())
            topology.topology_update(changes)
            assert topology.get_attributes("device.divvy") is not None

            # 2. Instance gone device
            dev = topology.get_device("divvy")
            assert dev.status is not ProxyStatus.OFFLINE
            gone_hash = Hash("device", Hash("divvy", None))
            changes = Hash("new", Hash(), "update", Hash(), "gone", gone_hash)
            topology.topology_update(changes)
            assert topology.get_attributes("device.divvy") is None
            # 2.1 A second time to be sure
            topology.topology_update(changes)
            assert topology.get_attributes("device.divvy") is None
            assert dev.status is ProxyStatus.OFFLINE

            # 3. instance gone macro
            assert topology.get_attributes("macro.macdonald") is not None
            gone_hash = Hash("macro", Hash("macdonald", None))
            changes = Hash("new", Hash(), "update", Hash(), "gone", gone_hash)
            topology.topology_update(changes)
            assert topology.get_attributes("macro.macdonald") is None
            # 3.1 A second time to be sure
            topology.topology_update(changes)
            assert topology.get_attributes("macro.macdonald") is None

            # 4. instance gone server
            schema = FooClass.getClassSchema()
            topology.class_schema_updated("swerver", "FooClass", schema)
            schema = topology.get_schema("swerver", "FooClass")
            assert schema is not None
            assert topology.get_attributes("server.swerver") is not None
            gone_hash = Hash("server", Hash("swerver", None))
            changes = Hash("new", Hash(), "update", Hash(), "gone", gone_hash)
            topology.topology_update(changes)
            assert topology.get_attributes("server.swerver") is None
            # 3.1 A second time to be sure
            topology.topology_update(changes)
            assert topology.get_attributes("server.swerver") is None
            assert "swerver" in topology._class_schemas
            # All class schemas are erased
            classes = topology._class_schemas["swerver"]
            assert not len(classes)

            # 5. Instance gone client
            assert topology.get_attributes("client.charlie") is not None
            gone_hash = Hash("client", Hash("charlie", None))
            changes = Hash("new", Hash(), "update", Hash(), "gone", gone_hash)
            topology.topology_update(changes)
            assert topology.get_attributes("client.charlie") is None
            # 4.1 A second time, stress
            topology.topology_update(changes)
            assert topology.get_attributes("client.charlie") is None

    def test_system_topology_update_instance(self):
        network = Mock()
        topology = SystemTopology()
        topology.initialize(system_hash())
        with singletons(network=network, topology=topology):
            attrs = topology.get_attributes("device.divvy")
            assert attrs["status"] != "error"

            h = Hash()
            h["device.divvy"] = None
            h["device.divvy", ...] = {
                "host": "BIG_IRON",
                "serverId": "swerver",
                "classId": "FooClass",
                "status": "error",
                "capabilities": 0,
            }

            dev = topology.get_device("divvy")

            assert dev.status is ProxyStatus.ONLINEREQUESTED
            network.onGetDeviceSchema.assert_called_with("divvy")
            assert dev.topology_node.status is not InstanceStatus.ERROR

            # Instance update device
            changes = Hash("new", Hash(), "update", h, "gone", Hash())
            topology.topology_update(changes)
            attrs = topology.get_attributes("device.divvy")
            assert attrs["status"] == "error"
            assert dev.topology_node.status is InstanceStatus.ERROR

            # A second time to be sure
            h["device.divvy", "status"] = "ok"
            changes = Hash("new", Hash(), "update", h, "gone", Hash())
            topology.topology_update(changes)
            attrs = topology.get_attributes("device.divvy")
            assert attrs["status"] == "ok"
            assert dev.topology_node.status is InstanceStatus.OK

    def test_visit_system_topology(self):
        topology = SystemTopology()
        topology.initialize(system_hash())
        with singletons(topology=topology):
            devices = []

            def device_visitor(node):
                nonlocal devices
                # Check for devices, level 3
                if node.level == 3:
                    devices.append(node)

            topology.visit_system_tree(device_visitor)
            assert len(devices) == 2
            assert devices[0].node_id == "divvy"
            assert devices[1].node_id == "macdonald"

            servers = []

            def server_visitor(node):
                nonlocal servers
                # Check for servers, level 1
                if node.level == 1:
                    servers.append(node)

            topology.visit_system_tree(server_visitor)
            assert len(servers) == 1
            assert servers[0].node_id == "swerver"

            # Visit the device tree
            # The system tree is a bit more strict and validates the Karaob
            # naming convention. Hence, we do not expect any devices

            members = []

            def member_visitor(node):
                nonlocal members
                # Check for member, level 2
                if node.level == 2:
                    members.append(node)

            topology.visit_device_tree(member_visitor)
            assert len(members) == 0

            h = Hash()
            h["device.EG/RR/VIEW"] = None
            h["device.EG/RR/VIEW", ...] = {
                "host": "BIG_IRON",
                "serverId": "swerver",
                "classId": "FooClass",
                "status": "ok",
                "visibility": 4,
                "capabilities": 0,
                "interfaces": 0,
            }

            changes = Hash("new", h, "update", Hash(), "gone", Hash())
            topology.topology_update(changes)

            topology.visit_device_tree(member_visitor)
            assert len(members) == 1
            assert members[0].node_id == "EG/RR/VIEW"

            assert is_device_online("EG/RR/VIEW")

    def test_topology_utils(self):
        topology = SystemTopology()
        topology.initialize(system_hash())
        with singletons(topology=topology):
            assert is_server_online("swerver")
            assert not is_server_online("cppserver/notthere")
            assert not get_macro_servers()

            h = Hash()
            h["server.karabo/macroServer"] = None
            h["server.karabo/macroServer", ...] = {
                "host": "BIG_IRON",
                "deviceClasses": ["Macro", "MetaMacro"],
                "lang": "macro",
                "type": "server",
                "serverId": "karabo/macroServer",
                "heartbeatInterval": 20,
                "karaboVersion": "2.13.0",
                "visibility": 4,
                "log": "INFO",
                "visibilities": [4, 4],
            }

            # adding an example of a macro started by a python server
            h["server.karabo/middlelayer"] = None
            h["server.karabo/middlelayer", ...] = {
                "host": "BIG_IRON",
                "lang": "python",
                "type": "server",
                "serverId": "karabo/middlelayer",
                "heartbeatInterval": 20,
                "karaboVersion": "2.13.0",
                "visibility": 4,
                "log": "INFO",
                "visibilities": [4, 4],
            }
            h["device.rogue_macro"] = None
            h["device.rogue_macro", ...] = {
                "host": "BIG_IRON",
                "classId": "RogueClass",
                "type": "macro",
                "serverId": "karabo/middlelayer",
                "heartbeatInterval": 120,
                "karaboVersion": "2.13.0",
                "visibility": 4,
                "status": "ok",
                "capabilities": 0,
            }

            changes = Hash("new", h, "update", Hash(), "gone", Hash())
            topology.topology_update(changes)
            servers = get_macro_servers()
            assert len(servers) == 1
            assert servers[0] == "karabo/macroServer"

            h = Hash()
            h["server.karabo/macroServer1"] = None
            h["server.karabo/macroServer1", ...] = {
                "host": "BIG_IRON",
                "deviceClasses": ["Macro", "MetaMacro"],
                "lang": "macro",
                "type": "server",
                "serverId": "karabo/macroServer1",
                "heartbeatInterval": 20,
                "karaboVersion": "2.13.0",
                "visibility": 4,
                "log": "INFO",
                "serverFlags": 1,
                "visibilities": [4, 4],
            }
            changes = Hash("new", h, "update", Hash(), "gone", Hash())
            topology.topology_update(changes)

            servers = get_macro_servers(False)
            assert len(servers) == 1
            servers = get_macro_servers(True)
            assert len(servers) == 2
            assert servers[1] == "karabo/macroServer1"

            # And one more macro server
            h = Hash()
            h["server.karabo/macroServer2"] = None
            h["server.karabo/macroServer2", ...] = {
                "host": "BIG_IRON",
                "deviceClasses": ["Macro", "MetaMacro"],
                "lang": "macro",
                "type": "server",
                "serverId": "karabo/macroServer2",
                "heartbeatInterval": 20,
                "karaboVersion": "2.20.0",
                "visibility": 4,
                "log": "INFO",
                "visibilities": [4, 4],
            }
            changes = Hash("new", h, "update", Hash(), "gone", Hash())
            topology.topology_update(changes)
            servers = get_macro_servers(False)
            assert len(servers) == 2
            # Least loaded
            assert servers[0] == "karabo/macroServer"

            # add one more macro
            h["macro.macromotor"] = None
            h["macro.macromotor", ...] = {
                "host": "BIG_IRON",
                "classId": "RogueClass",
                "type": "macro",
                "serverId": "karabo/macroServer",
                "heartbeatInterval": 120,
                "karaboVersion": "2.13.0",
                "visibility": 4,
                "status": "ok",
                "capabilities": 0,
            }
            changes = Hash("new", h, "update", Hash(), "gone", Hash())
            topology.topology_update(changes)
            servers = get_macro_servers(False)
            assert len(servers) == 2
            # Least loaded changed
            assert servers[0] == "karabo/macroServer2"



# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.

from qtpy import uic
from qtpy.QtCore import QPoint, QSize, Qt, Slot
from qtpy.QtGui import QStandardItem, QStandardItemModel
from qtpy.QtWidgets import (
    QAbstractItemView, QDialog, QHeaderView, QMenu, QTreeView)

from karabo.common.api import ServerFlags
from karabogui import icons
from karabogui.dialogs.utils import get_dialog_ui
from karabogui.navigation.utils import get_language_icon
from karabogui.singletons.api import get_manager, get_topology

HEADER_LABELS = ["Server Id", "Host"]


class DevelopmentTopologyView(QTreeView):
    def __init__(self, parent=None):
        super().__init__(parent)
        model = QStandardItemModel(parent=self)
        model.setHorizontalHeaderLabels(HEADER_LABELS)
        self.setModel(model)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        header = self.header()
        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        self.setUniformRowHeights(True)

    @Slot()
    def refresh(self):
        model = self.model()
        model.clear()
        model.setHorizontalHeaderLabels(HEADER_LABELS)

        topology = get_topology()
        root_item = model.invisibleRootItem()

        def visitor(node):
            attributes = node.attributes

            if (attributes.get("serverFlags", 0) & ServerFlags.Development
                    == ServerFlags.Development):
                icon = get_language_icon(node)
                server_item = QStandardItem(icon, node.path)
                server_item.setEditable(False)

                host = attributes.get("host", "None")
                host_item = QStandardItem(host)
                host_item.setEditable(False)

                root_item.appendRow([server_item, host_item])

        topology.visit_system_tree(visitor)


class DevelopmentTopologyDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        ui_file = get_dialog_ui("development_topology.ui")
        uic.loadUi(ui_file, self)

        self.tree_view = DevelopmentTopologyView(parent=self)

        self.ui_refresh.setIcon(icons.refresh)
        self.ui_refresh.clicked.connect(self.tree_view.refresh)

        layout = self.ui_frame
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.tree_view)

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._context_menu)

        self.tree_view.refresh()

    @Slot(QPoint)
    def _context_menu(self, pos):
        index = self.tree_view.selectionModel().currentIndex()
        if not index.isValid():
            return

        menu = QMenu(parent=self)
        shutdown_action = menu.addAction('Shutdown')
        shutdown_action.triggered.connect(self.onKillInstance)
        shutdown_action.setIcon(icons.delete)

        menu.exec(self.mapToGlobal(pos))

    @Slot()
    def onKillInstance(self):
        manager = get_manager()
        index = self.tree_view.selectionModel().currentIndex()
        serverId = index.sibling(index.row(), 0).data()
        manager.shutdownServer(serverId, parent=self)

    def sizeHint(self):
        return QSize(640, 480)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from enum import Enum, IntEnum


class Axes(IntEnum):
    Y = 1
    X = 0
    Z = 2


class AxisType(Enum):
    AuxPlot = "AuxPlots"
    Alarm = "Alarm"
    Classic = "Classic"
    State = "State"
    Time = "Time"


class MouseTool(IntEnum):
    # Default
    Pointer = 0
    Zoom = 1
    Move = 2

    # Optional
    Picker = 3


# Backward compatibility
MouseMode = MouseTool


class ExportTool(Enum):
    Image = "All items"
    Data = "npy"


class AuxPlots(IntEnum):
    NoPlot = 0
    ProfilePlot = 1
    Histogram = 2


class ROITool(IntEnum):
    NoROI = 0
    Rect = 1
    Crosshair = 2

    DrawRect = 11
    DrawCrosshair = 12


class AspectRatio(IntEnum):
    NoAspectRatio = 0
    PixelDependent = 1
    ScaleDependent = 2



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import mock

from qtpy.QtCore import QSize
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.api import SceneModel
from karabogui.singletons.mediator import Mediator
from karabogui.testing import GuiTestCase, singletons

from ..scenepanel import ScenePanel

DIALOG_PATH = "karabogui.panels.scenepanel.ResizeSceneDialog"


class TestScenePanel(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.model = SceneModel()
        self.mediator = Mediator()
        # Use own mediator as closure will send signals
        with singletons(mediator=self.mediator):
            self.panel = ScenePanel(self.model, connected_to_server=True)
            container = mock.MagicMock()
            container.maximized.new = False
            self.panel.attach_to_container(container)
            self.panel.show()

    def tearDown(self):
        super().tearDown()
        with singletons(mediator=self.mediator):
            self.panel.close()
            self.panel.destroy()
            self.model = None
            self.panel = None
        self.mediator = None

    def test_resize_scene(self):
        panel_size = self.panel.size()
        assert self.scene_size == self.model_size

        # Docked
        width, height = (100, 200)
        self._resize_scene(width, height, docked=True)
        assert self.panel.is_docked
        assert self.scene_size == (width, height)
        assert self.scene_size == self.model_size
        assert self.panel.size() == panel_size

        # Undocked
        width, height = (300, 400)
        self.panel.onUndock()
        self._resize_scene(width, height, docked=False)
        assert not self.panel.is_docked
        assert self.scene_size == (width, height)
        assert self.scene_size == self.model_size
        assert self.panel.size() != panel_size

    def _resize_scene(self, width=100, height=200, docked=False):
        mocked_exec = mock.patch(f"{DIALOG_PATH}.exec",
                                 return_value=QDialog.Accepted)
        mocked_size = mock.patch(f"{DIALOG_PATH}.scene_size",
                                 new=QSize(width, height))
        with mocked_exec, mocked_size:
            self.panel.design_mode_changed(True)
            self.panel.is_docked = docked
            self.panel.resize_scene()

    @property
    def scene_size(self):
        size = self.panel.scene_view.size()
        return size.width(), size.height()

    @property
    def model_size(self):
        return self.model.width, self.model.height



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os.path as op

from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QLabel
from traits.api import Instance

from karabo.common.api import State
from karabo.common.scenemodel.api import LampModel
from karabogui import icons
from karabogui.binding.api import StringBinding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller, with_display_type)

ICONS_DIR = op.dirname(icons.__file__)
STATE_LAMP_PATH = {
    State.CHANGING: op.join(ICONS_DIR, 'lamp-changing'),
    State.RUNNING: op.join(ICONS_DIR, 'lamp-running'),
    State.ACTIVE: op.join(ICONS_DIR, 'lamp-active'),
    State.PASSIVE: op.join(ICONS_DIR, 'lamp-passive'),
    State.STATIC: op.join(ICONS_DIR, 'lamp-static'),
    State.INIT: op.join(ICONS_DIR, 'lamp-init'),
    State.NORMAL: op.join(ICONS_DIR, 'lamp-known'),
    State.KNOWN: op.join(ICONS_DIR, 'lamp-known'),
    State.ERROR: op.join(ICONS_DIR, 'lamp-error'),
    State.UNKNOWN: op.join(ICONS_DIR, 'lamp-unknown'),
    State.DISABLED: op.join(ICONS_DIR, 'lamp-disabled')
}


@register_binding_controller(ui_name='Generic Lamp', klassname='LampWidget',
                             binding_type=StringBinding,
                             is_compatible=with_display_type('State'),
                             can_show_nothing=False)
class LampWidget(BaseBindingController):
    model = Instance(LampModel, args=())

    def create_widget(self, parent):
        widget = QLabel(parent)
        widget.setScaledContents(True)
        return widget

    def value_update(self, proxy):
        value = proxy.value
        state = State(value)
        if state.isDerivedFrom(State.CHANGING):
            self._set_lamp(STATE_LAMP_PATH[State.CHANGING])
        elif state.isDerivedFrom(State.RUNNING):
            self._set_lamp(STATE_LAMP_PATH[State.RUNNING])
        elif state.isDerivedFrom(State.ACTIVE):
            self._set_lamp(STATE_LAMP_PATH[State.ACTIVE])
        elif state.isDerivedFrom(State.PASSIVE):
            self._set_lamp(STATE_LAMP_PATH[State.PASSIVE])
        elif state.isDerivedFrom(State.DISABLED):
            self._set_lamp(STATE_LAMP_PATH[State.DISABLED])
        elif state is State.STATIC:
            self._set_lamp(STATE_LAMP_PATH[State.STATIC])
        elif state is State.NORMAL:
            self._set_lamp(STATE_LAMP_PATH[State.NORMAL])
        elif state is State.ERROR:
            self._set_lamp(STATE_LAMP_PATH[State.ERROR])
        elif state is State.INIT:
            self._set_lamp(STATE_LAMP_PATH[State.INIT])
        else:
            self._set_lamp(STATE_LAMP_PATH[State.UNKNOWN])

    def _set_lamp(self, path):
        p = QPixmap(path)
        self.widget.setPixmap(p)
        self.widget.setMaximumWidth(p.width())
        self.widget.setMaximumHeight(p.height())



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QStyle

from karabo.native import Hash
from karabogui.binding.api import ProxyStatus
from karabogui.dialogs.api import ProjectTopologyDialog
from karabogui.testing import singletons, system_hash
from karabogui.topology.api import SystemTopology


def test_project_topology_dialog(gui_app):
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(topology=topology):
        dialog = ProjectTopologyDialog()

        devices = [Hash(
            "device_name", "divvy",
            "device_class", "Fooclass",
            "server_name", "swerver",
            "device_uuid", "test-uuid",
            "project_name", "KaraboRocks")]

        h = Hash("items", devices)
        dialog.initialize(h)
        assert dialog.table_model.rowCount() == 1
        assert dialog.table_model.columnCount() == 6

        assert dialog.filter_model.rowCount() == 1
        assert dialog.table_model.index(0, 0).data(
            Qt.UserRole) == ProxyStatus.NOPLUGIN
        assert dialog.table_model.index(0, 1).data(Qt.UserRole) == "divvy"
        assert dialog.table_model.index(0, 1).data(Qt.DisplayRole) == "divvy"

        index = dialog.table_model.index(0, 1)
        state = dialog.delegate.button_state(index)
        assert state == QStyle.State_On



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from functools import partial

from pyqtgraph import GraphicsObject, GraphicsView, PlotItem, mkBrush, mkPen
from qtpy.QtCore import QSize, Signal, Slot
from qtpy.QtGui import QColor, QPalette
from qtpy.QtWidgets import QAction, QGridLayout, QSizePolicy, QWidget

from karabo.common.scenemodel.api import CurveType
from karabogui import icons
from karabogui.actions import KaraboAction, build_qaction
from karabogui.graph.common.api import (
    AxesLabelsDialog, AxisType, BaseROIController, ExportTool, ImageExporter,
    KaraboLegend, KaraboViewBox, MouseTool, PlotDataExporter, PointCanvas,
    ROITool, ToolbarController, create_axis_items, get_default_brush,
    get_default_pen, make_pen, safe_log10)
from karabogui.graph.common.const import (
    ACTION_ITEMS, AXIS_ITEMS, CHECK_ACTIONS, DEFAULT_BAR_WIDTH, DEFAULT_SYMBOL,
    EMPTY_SYMBOL_OPTIONS, SYMBOL_SIZE, WIDGET_HEIGHT_HINT, WIDGET_MIN_HEIGHT,
    WIDGET_MIN_WIDTH, WIDGET_WIDTH_HINT)
from karabogui.graph.plots.dialogs import GraphViewDialog, RangeDialog
from karabogui.graph.plots.items import ScatterGraphPlot, VectorBarGraphPlot
from karabogui.graph.plots.tools import CrossTargetController

ALPHA_GRID = 30 / 255
TICK_AXES = ["bottom", "left"]
ROI_CANVAS_MAP = {
    ROITool.DrawCrosshair: (PointCanvas, ROITool.Crosshair)}


class KaraboPlotView(QWidget):
    """ The main KaraboPlotView widget for ImagePlots

    This widget houses the plotItem and the toolbar for other tooling features
    """
    stateChanged = Signal(object)

    def __init__(self, axis=AxisType.Classic, actions=None, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setMinimumSize(WIDGET_MIN_WIDTH, WIDGET_MIN_HEIGHT)

        # Main layout to organize
        layout = QGridLayout(self)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        # Initialize axis items
        axis_items = create_axis_items(axis, axes_with_ticks=TICK_AXES)

        # Our viewbox with reset range addition!
        viewbox = KaraboViewBox(parent=None)
        viewbox.menu.addAction("Reset Range", self.reset_range)

        # Add our basic plotItem to this widget
        self.plotItem = PlotItem(viewBox=viewbox, axisItems=axis_items,
                                 enableMenu=False, parent=None)

        # Improve axes rendering
        for axis in AXIS_ITEMS:
            axis_item = self.plotItem.getAxis(axis)
            axis_item.setZValue(0)

        self.graph_view = GraphicsView(parent=self)
        self.graph_view.setAntialiasing(False)
        self.graph_view.enableMouse(False)
        # Erase all transparent palettes (Performance fix: PyQtGraph == 0.11.1)
        self.graph_view.setPalette(QPalette())

        # row, col, row_span, col_span
        self.layout().addWidget(self.graph_view, 0, 0, 1, 1)
        # PlotItem gets added automatically!
        self.graph_view.setCentralItem(self.plotItem)

        self.show_all_axis()

        self.configuration = {}
        self.qactions = {}
        actions = actions or ACTION_ITEMS
        self.setup_qactions(actions)

        # Our tooling instances
        self._toolbar = None
        self._legend = None
        self._cross_target = None
        self._roi = None
        self._canvas = None

        self._show_symbols = False

        self._data_point_toggle_enabled = False
        self._show_legend_action = None

    def sizeHint(self):
        """Provide a size hint slightly larger than the minimum size"""
        return QSize(WIDGET_WIDTH_HINT, WIDGET_HEIGHT_HINT)

    def setup_qactions(self, actions):
        """The setup of the basic actions is gathered here!"""

        x_grid = KaraboAction(
            text="Grid X",
            tooltip="Toggle the grid of the X-Axis",
            triggered=self.set_grid_x,
            checkable=True,
            name='x_grid')

        y_grid = KaraboAction(
            text="Grid Y",
            tooltip="Toggle the grid of the Y-Axis",
            triggered=self.set_grid_y,
            checkable=True,
            name='y_grid')

        x_log = KaraboAction(
            text="Log X",
            tooltip="Toggle the logarithmic scale of the X-Axis",
            triggered=self.set_log_x,
            checkable=True,
            name='x_log')

        y_log = KaraboAction(
            text="Log Y",
            tooltip="Toggle the logarithmic scale of the Y-Axis",
            triggered=self.set_log_y,
            checkable=True,
            name='y_log')

        x_invert = KaraboAction(
            text="Invert X",
            tooltip="Invert the scale of the X-Axis",
            triggered=self.set_invert_x,
            checkable=True,
            name='x_invert')

        y_invert = KaraboAction(
            text="Invert Y",
            tooltip="Invert the scale of the Y-Axis",
            triggered=self.set_invert_y,
            checkable=True,
            name='y_invert')

        x_range = KaraboAction(
            text="Range X",
            tooltip="Configure the range of the X-Axis",
            triggered=self.configure_range_x,
            name='x_range')

        y_range = KaraboAction(
            text="Range Y",
            tooltip="Configure the range of the Y-Axis",
            triggered=self.configure_range_y,
            name='y_range')

        axes = KaraboAction(
            text="Axes",
            tooltip="Configure the axes labels and units",
            triggered=self.configure_axes,
            name='axes')

        view = KaraboAction(
            text="View",
            tooltip="Configure the background and title of the graph view",
            triggered=self.configure_view,
            name='view')

        for check_action in (x_grid, y_grid, x_log, y_log, x_invert, y_invert):
            if check_action.name in actions:
                q_ac = build_qaction(check_action, self)
                q_ac.triggered.connect(partial(self.check_action_callback,
                                               callback=check_action.triggered,
                                               name=check_action.name))
                self.qactions[check_action.name] = q_ac
                self.addAction(q_ac)

        viewbox = self.plotItem.vb
        for action in (x_log, y_log):
            if action.name in self.qactions:
                viewbox.add_action(self.qactions[action.name], separator=False)

        for k_action in (axes, x_range, y_range, view):
            if k_action.name in actions:
                q_ac = build_qaction(k_action, self)
                q_ac.triggered.connect(k_action.triggered)
                self.addAction(q_ac)

        if x_range.name in actions:
            axis_item = self.plotItem.getAxis('bottom')
            axis_item.axisDoubleClicked.connect(self.configure_range_x)

        if y_range.name in actions:
            axis_item = self.plotItem.getAxis('left')
            axis_item.axisDoubleClicked.connect(self.configure_range_y)

    # ----------------------------------------------------------------
    # Base Action Slots

    @Slot(object, object, object)
    def check_action_callback(self, state, callback=None, name=None):
        if callback is not None:
            callback(state)

        self.qactions[name].setChecked(state)
        self.configuration[name] = state
        self.stateChanged.emit({name: state})

    def set_grid_x(self, state):
        alpha_value = ALPHA_GRID if state else None
        self.plotItem.showGrid(x=state, alpha=alpha_value)

    def set_grid_y(self, state):
        alpha_value = ALPHA_GRID if state else None
        self.plotItem.showGrid(y=state, alpha=alpha_value)

    def set_log_x(self, state):
        self.plotItem.setLogMode(x=state)

    def set_log_y(self, state):
        self.plotItem.setLogMode(y=state)

    def set_invert_x(self, state):
        self.plotItem.invertX(state)

    def set_invert_y(self, state):
        self.plotItem.invertY(state)

    @Slot()
    def configure_axes(self):
        config, ok = AxesLabelsDialog.get(self.configuration, parent=self)
        if ok:
            self.set_label(axis=0, text=config['x_label'],
                           units=config['x_units'])
            self.set_label(axis=1, text=config['y_label'],
                           units=config['y_units'])
            self.configuration.update(**config)
            self.stateChanged.emit(config)

    @Slot()
    def configure_range_x(self):
        x_min, x_max = self.get_view_range_x()
        actual = {'x_min': x_min, 'x_max': x_max}
        config, ok = RangeDialog.get(self.configuration, actual, axis=0,
                                     parent=self)
        if ok:
            self.configuration.update(**config)
            x_autorange = config['x_autorange']
            if not x_autorange:
                self.set_range_x(*self.get_config_range_x())
            else:
                self.set_autorange_x(x_autorange)
            self.stateChanged.emit(config)

    @Slot()
    def configure_range_y(self):
        y_min, y_max = self.get_view_range_y()
        actual = {'y_min': y_min, 'y_max': y_max}
        config, ok = RangeDialog.get(self.configuration, actual, axis=1,
                                     parent=self)
        if ok:
            self.configuration.update(**config)
            autorange = config['y_autorange']
            if not autorange:
                self.set_range_y(*self.get_config_range_y())
            else:
                self.set_autorange_y(autorange)
            self.stateChanged.emit(config)

    @Slot()
    def configure_view(self):
        config, ok = GraphViewDialog.get(self.configuration, parent=self)
        if ok:
            self.set_title(config['title'])
            self.set_background(config['background'])
            self.configuration.update(**config)
            self.stateChanged.emit(config)

    @Slot()
    def reset_range(self):
        self.reset_range_x()
        self.reset_range_y()

    @Slot(bool)
    def toggle_data_symbols(self, show):
        """Toggle the data points on the plotItems"""
        self._show_symbols = show
        for item in self.plotItem.dataItems[:]:
            if show:
                pen = item.opts['pen']
                options = {'symbol': DEFAULT_SYMBOL,
                           'symbolSize': SYMBOL_SIZE,
                           'symbolPen': pen,
                           'symbolBrush': mkBrush(pen.color())}
            else:
                options = EMPTY_SYMBOL_OPTIONS
            # NOTE: We are directly setting the options dict to avoid
            # multiple updates!
            item.opts.update(options)
            item.updateItems()

    # ----------------------------------------------------------------
    # Base Functions

    def set_background(self, color):
        """Set the background color of the overall base plot widget"""
        self.graph_view.setBackground(QColor(color))
        if self._toolbar is not None:
            self._toolbar.set_background(QColor(color))

    def reset_range_x(self):
        config = self.configuration
        x_autorange = config['x_autorange']
        if not x_autorange:
            self.set_range_x(*self.get_config_range_x())
        else:
            self.set_autorange_x(x_autorange)

    def reset_range_y(self):
        config = self.configuration
        y_autorange = config['y_autorange']
        if not y_autorange:
            self.set_range_y(*self.get_config_range_y())
        else:
            self.set_autorange_y(y_autorange)

    def enable_data_toggle(self, activate=False):
        """Optionally enable the showing of data points"""
        toggle_action = QAction("Show data points", self)
        toggle_action.setCheckable(True)
        toggle_action.setChecked(False)
        toggle_action.toggled.connect(self.toggle_data_symbols)
        viewbox = self.plotItem.vb
        seperator = self._legend is None
        viewbox.add_action(toggle_action, separator=seperator)
        if activate:
            toggle_action.setChecked(True)
        self._data_point_toggle_enabled = True

    def enable_export(self):
        """Optionally enable the showing of data points"""
        # Add Export toolset
        if self._toolbar is not None:
            export_toolset = self._toolbar.add_toolset(ExportTool)
            export_toolset.on_trait_change(self.export, "current_tool")

    def restore(self, config):
        """Restore the widget configuration with a config dictionary"""
        self.configuration.update(**config)
        actions = [name for name in CHECK_ACTIONS if name in self.qactions]
        for name in actions:
            self.qactions[name].setChecked(self.configuration[name])

        # Set x and y grid at the same time to avoid overwritten settings
        self.plotItem.showGrid(config['x_grid'], config['y_grid'],
                               alpha=ALPHA_GRID)

        self.set_log_x(config['x_log'])
        self.set_log_y(config['y_log'])
        self.set_label(axis=0, text=config['x_label'], units=config['x_units'])
        self.set_label(axis=1, text=config['y_label'], units=config['y_units'])
        self.set_invert_x(config['x_invert'])
        self.set_invert_y(config['y_invert'])

        # Restore ROIs
        if self._roi is not None:
            self._roi.remove_all()
            roi_tool = config['roi_tool']
            roi_items = config.get('roi_items', [])
            for roi_data in roi_items:
                roi_type = roi_data['roi_type']
                name = roi_data['name']
                if roi_type == ROITool.Crosshair:
                    pos = (roi_data['x'], roi_data['y'])
                    size = None
                else:
                    continue

                self._roi.add(roi_type, pos, size=size, name=name)

            self._roi.show(roi_tool)

            # Make sure that we checked the tool!
            if self._toolbar is not None:
                if roi_tool != ROITool.NoROI:
                    tool = ROITool(roi_tool).name
                    self._toolbar.buttons[tool].setChecked(True)

        self.reset_range()

        # Set the view configuration! Some widgets, e.g. external extensions
        # might not define `background` and `title`.
        if 'background' in config:
            self.set_background(config['background'])
        if 'title' in config:
            self.set_title(config['title'])

    def apply_curve_options(self, options: dict):
        """Apply the plotting options to the corresponding curves"""
        for plot_item, curve_opts in options.items():
            item_options = plot_item.opts

            curve_type = curve_opts.get("curve_type")
            assert curve_type in {CurveType.Curve, CurveType.Trend}

            name = curve_opts.get("name")
            item_options["name"] = name

            pen_color = curve_opts.get("pen_color")
            pen = mkPen(pen_color)
            item_options["pen"] = pen

            if self._show_symbols:
                item_options["symbolBrush"] = mkBrush(pen.color())
                item_options["symbolPen"] = pen

            for sample, label in self._legend.items:
                if sample.item is plot_item:
                    sample.setBrush(mkBrush(pen.color()))
                    sample.setPen(pen)
                    label.setText(name)
                    break

            self._legend.updateSize()
            plot_item.updateItems()

    # ----------------------------------------------------------------
    # Toolbar functions Events

    def add_cross_target(self):
        if self._cross_target is None:
            self._cross_target = CrossTargetController(self.plotItem)

        return self._cross_target

    def add_roi(self, enable=True):
        if enable and self._roi is None:
            # Initialize ROI controller
            self._roi = BaseROIController(self.plotItem)
            self._roi.set_pen(make_pen('k'))
            set_roi = KaraboAction(
                text="Set ROI",
                icon=icons.apply,
                tooltip="Set the ROI information",
                triggered=self._set_roi_configuration)
            q_ac = build_qaction(set_roi, self)
            q_ac.triggered.connect(set_roi.triggered)
            self.addAction(q_ac)

        elif not enable and self._roi is not None:
            # Remove ROI items and destroy controller
            self._roi.remove_all()
            self._roi.destroy()
            self._roi = None

        return self._roi

    @Slot()
    def _set_roi_configuration(self):
        """Set ROI information to the model"""
        items = []
        for tool, roi_items in self._roi.roi_items.items():
            # Each tool has multiple roi objects
            for roi in roi_items:
                traits = {'roi_type': tool, 'name': roi.name}
                if tool == ROITool.Crosshair:
                    x, y = roi.coords
                    traits.update({'x': x, 'y': y})
                    items.append(traits)
        config = {}
        config['roi_tool'] = self._roi.current_tool
        config['roi_items'] = items
        self.configuration.update(**config)
        self.stateChanged.emit(config)

    def add_toolbar(self):
        """Enable the standard toolbar on the widget

        This is the finalizer of the `KaraboPlotView` widget. All signals
        should be connected here!
        """
        if self._toolbar is not None:
            raise ValueError('Toolbar already added')

        # Instantiate toolbar, has a default mouse mode toolset
        self._toolbar = tb = ToolbarController(parent=self)
        tb.toolsets[MouseTool].on_trait_change(
            self.plotItem.vb.set_mouse_tool,
            "current_tool")

        # Add ROI toolset
        if self._roi is not None:
            crosshair_button = tb.add_tool(ROITool.Crosshair)
            crosshair_button.triggered.connect(self._activate_roi_tool)

            # Connect signals from the items
            self._roi.selected.connect(crosshair_button.check)
            uncheck_button = partial(crosshair_button.check, None)
            self._roi.removed.connect(uncheck_button)

        # Add Crosshair toolset
        if self._cross_target is not None:
            tb.add_button(self._cross_target.action_button)

        # row, col, row_span, col_span
        self.layout().addWidget(tb.widget, 0, 1, 1, 1)

        return self._toolbar

    @Slot(object)
    def export(self, export_type=ExportTool.Data):
        """Exports the data or image according to the desired format"""
        if export_type == ExportTool.Image:
            scene = self.plotItem.scene()
            exporter = ImageExporter(scene)
        elif export_type == ExportTool.Data:
            if not len(self.plotItem.dataItems):
                return
            exporter = PlotDataExporter(self.plotItem.dataItems)

        exporter.export()

    # ----------------------------------------------------------------
    # PlotItem methods

    def clear(self):
        """Remove all the items in the PlotItem object"""
        plot_item = self.plotItem
        for item in plot_item.items[:]:
            plot_item.removeItem(item)

    def clearData(self):
        plot_item = self.plotItem
        for item in plot_item.dataItems[:]:
            item.clear()

    def add_curve_fill(self, name=None, pen=None, brush=None, **options):
        if pen is None:
            pen = get_default_pen()
        if brush is None:
            brush = get_default_brush()
        item = self.plotItem.plot(name=name, pen=pen, fillLevel=0,
                                  fillBrush=brush, **options)
        item.setDynamicRangeLimit(None)
        return item

    def add_curve_item(self, name=None, pen=None, **options):
        """Add a plot to the built-in plotItem from PyQtGraph

        :param name: Set a name to automatically provide a legend

        :param pen: Pen to use for drawing line between points.
                    Default is solid grey, 1px width. Use None to disable line
                    drawing.
        """
        if pen is None:
            pen = get_default_pen()

        if self._show_symbols:
            # If a curve is dynamically added, we still have to show points
            # when desired!
            options.update({'symbol': DEFAULT_SYMBOL,
                            'symbolSize': SYMBOL_SIZE,
                            'symbolPen': pen,
                            'symbolBrush': mkBrush(pen.color())})

        item = self.plotItem.plot(name=name, pen=pen, **options)
        item.setDynamicRangeLimit(None)
        return item

    def add_scatter_item(self, pen=None, cycle=True, **options):
        """Add a scatter item to the built-in plotItem

        :param pen: pen used for the scatter plot (default: None)
        :param cycle: Defines whether the latest brush color is different

        Optionally as keyword arguments:

        :parem size: The size (or list of sizes) of spots.
        """
        if pen is None:
            pen = mkPen(None)

        item = ScatterGraphPlot(pen=pen, cycle=cycle, **options)
        self.plotItem.addItem(item)

        return item

    def add_bar_item(self, width=DEFAULT_BAR_WIDTH, brush=None):
        """Add a bar item to the built-in plotItem

        :param width: width used for the bar plot
        :param brush: brush to be used when plotting (default: fancy blue)
        """
        if brush is None:
            brush = get_default_brush()
        item = VectorBarGraphPlot(width, brush)
        self.plotItem.addItem(item)

        return item

    def remove_item(self, item):
        """Remove any plot item to the built-in plotItem"""
        self.plotItem.removeItem(item)

    def set_aspect_locked(self, enable, ratio=1):
        self.plotItem.setAspectLocked(enable, ratio)

    # ----------------------------------------------------------------
    # Legend methods

    def add_legend(self, size=None, offset=(5, 5), visible=True):
        """Create a new LegendItem and anchor it over the internal ViewBox.

        Plots will be automatically displayed in the legend if they
        are created with the 'name' argument.

        :param size: the size of the legend (default: None)
        :param offset: the offset with default (30, 30)
        :param visible: Boolean to set whether this Legend should be visible
        """
        if self._legend is not None:
            return

        self._legend = KaraboLegend(size, offset)
        self._legend.setBrush(None)

        self.plotItem.legend = self._legend
        self._legend.setParentItem(self.plotItem.vb)
        if not visible:
            self.plotItem.legend.setVisible(False)

        # Add a menu-item to show/hide the legend
        show_legend_action = QAction("Show Legend", parent=self)
        show_legend_action.setCheckable(True)
        show_legend_action.setChecked(False)
        show_legend_action.triggered.connect(self._set_legend)
        seperator = not self._data_point_toggle_enabled
        self.plotItem.vb.add_action(show_legend_action, separator=seperator)
        self._show_legend_action = show_legend_action
        return self._legend

    def set_legend(self, visible: bool):
        """Set the legend item visible or invisible"""
        self._set_legend(visible)
        self._toggle_legend_menu(visible)

    @Slot(bool)
    def _set_legend(self, visible: bool):
        if self._legend is None:
            return
        self.plotItem.legend.setVisible(visible)

    def _toggle_legend_menu(self, toggled: bool):
        self._show_legend_action.setChecked(toggled)

    # ----------------------------------------------------------------
    # Axis and Title methods

    def set_label(self, axis, text="", units=""):
        """Set the label of an axis of the built-in plotItem

        :param axis: The axis, e.g. ``0`` for x-axis, ``1`` for y-axis
        :param text: The text used for the label (None)
        :param units: The declared unit string for the axis label (None)
        """
        self.plotItem.setLabel(TICK_AXES[axis], text, units)
        show = True if (text != "" or units != "") else False
        # XXX: We circumvent a PyQtGraph bug here and set the visibility
        # ourselves! Setting a label and removing it, gives PyQtGraph problems.
        self.plotItem.showLabel(TICK_AXES[axis], show=show)

    def set_title(self, title=None, **kwargs):
        """Set the title of the built-in plotItem"""
        title = f"<b>{title}</b>" if title else None
        self.plotItem.setTitle(title, **kwargs)

    def hide_all_axis(self):
        """Hide all axis of the widget"""
        for axis in AXIS_ITEMS:
            self.plotItem.hideAxis(axis)

    def show_all_axis(self):
        """Show all axis of the widget"""
        for axis in AXIS_ITEMS:
            self.plotItem.showAxis(axis)

    # ----------------------------------------------------------------
    # Range methods

    def set_autorange(self, enable):
        """Enable or disable the auto range for a specific axis"""
        if enable:
            self.plotItem.vb.enableAutoRange()
        else:
            self.plotItem.vb.disableAutoRange()

    def set_autorange_x(self, enable):
        """Enable or disable the auto range for a specific axis"""
        self.plotItem.vb.enableAutoRange(axis=0, enable=enable)

    def set_autorange_y(self, enable):
        """Enable or disable the auto range for a specific axis"""
        self.plotItem.vb.enableAutoRange(axis=1, enable=enable)

    def set_range_x(self, min_value, max_value):
        """Set the X Range of the view box with min and max value"""
        self.plotItem.vb.setRange(xRange=[min_value, max_value],
                                  disableAutoRange=True)

    def set_range_y(self, min_value, max_value):
        """Set the Y Range of the view box with min and max value"""
        self.plotItem.vb.setRange(yRange=[min_value, max_value],
                                  disableAutoRange=True)

    def get_view_range_x(self):
        view_range = self.plotItem.vb.viewRange()
        x_min, x_max = view_range[0]

        # Revert viewbox values (log scale) to linear scale
        if self.configuration['x_log']:
            x_min, x_max = 10 ** x_min, 10 ** x_max
        return x_min, x_max

    def get_view_range_y(self):
        view_range = self.plotItem.vb.viewRange()
        y_min, y_max = view_range[1]

        # Revert viewbox values (log scale) to linear scale
        if self.configuration['y_log']:
            y_min, y_max = 10 ** y_min, 10 ** y_max
        return y_min, y_max

    def get_config_range_x(self):
        """Returns the x-range from the configuration in viewbox coordinates
           (linear or log scale)"""
        config = self.configuration
        x_min, x_max = config['x_min'], config['x_max']
        # Revert input values (linear scale) to log scale if enabled
        if config['x_log']:
            x_min, x_max = safe_log10(x_min), safe_log10(x_max)
        return x_min, x_max

    def get_config_range_y(self):
        """Returns the y-range from the configuration in viewbox coordinates
           (linear or log scale)"""
        config = self.configuration
        y_min, y_max = config['y_min'], config['y_max']
        # Revert input values (linear scale) to log scale if enabled
        if config['y_log']:
            y_min, y_max = safe_log10(y_min), safe_log10(y_max)
        return y_min, y_max

    def update_legend_text(self, plot_item: GraphicsObject, text: str):
        """Set the label text for the  plot item's legend"""
        label = self._legend.getLabel(plot_item)
        if label is not None:
            label.setText(text)
            self._legend.updateSize()

    # -----------------------------------------------------------------------
    # ROI methods

    @Slot(object)
    def _activate_roi_tool(self, tool):
        if tool is None:
            tool = ROITool.NoROI

        if self._canvas is not None:
            self._deactivate_canvas()

        if tool in [ROITool.NoROI, ROITool.Crosshair]:
            self._roi.show(tool)
        elif tool is ROITool.DrawCrosshair:
            # Hide shown ROIs to give way to the canvas
            self._roi.show(ROITool.NoROI)
            self._activate_canvas(tool)

    def _activate_canvas(self, draw_tool):
        # Deactivate the canvas if a previous is applied
        if self._canvas is not None:
            self._deactivate_canvas()

        # Check if draw tool has an canvas counterpart
        if draw_tool not in ROI_CANVAS_MAP:
            return

        canvas_class, roi_tool = ROI_CANVAS_MAP[draw_tool]
        viewbox_rect = self.plotItem.vb.viewRect()

        # Create and add canvas
        self._canvas = canvas_class(viewbox_rect)
        self.plotItem.vb.addItem(self._canvas, ignoreBounds=True)
        self._canvas.editingFinished.connect(partial(self._draw_roi, roi_tool))

    @Slot(object)
    def _draw_roi(self, roi_tool, rect):
        self._deactivate_canvas()
        if rect.isValid():
            # Draw the ROI and show corresponding ROI tool
            self._roi.add(roi_tool, pos=rect.topLeft())
            self._roi.show(roi_tool)
            self._roi.selected.emit(roi_tool)
        else:
            # Do nothing and unselect the button
            self._roi.selected.emit(ROITool.NoROI)

    def _deactivate_canvas(self):
        self.plotItem.vb.removeItem(self._canvas)
        self._canvas.destroy()
        self._canvas = None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import QEvent, QPoint, Qt
from qtpy.QtGui import QDropEvent
from qtpy.QtWidgets import QWidget
from traits.api import Dict, Instance

from karabo.common.scenemodel.api import (
    BaseWidgetObjectData, BoxLayoutModel, SceneModel)
from karabo.common.states import State
from karabo.native import AccessMode, Configurable, Double, String
from karabogui.binding.api import FloatBinding, StringBinding
from karabogui.configurator.utils import dragged_configurator_items
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.sceneview.tools.scenedndhandler import ConfigurationDropHandler
from karabogui.sceneview.view import SceneView
from karabogui.testing import flushed_registry, get_property_proxy

GET_PROXY_PATH = "karabogui.sceneview.tools.scenedndhandler.get_proxy"


class TypeObject(Configurable):
    state = String(
        defaultValue=State.ON,
        enum=State,
        displayType="State",
        accessMode=AccessMode.READONLY)
    prop = Double(
        minInc=-1000.0,
        maxInc=1000.0,
        allowedStates=[State.ON])


class UniqueModel(BaseWidgetObjectData):
    attributes = Dict


@pytest.fixture(scope="module")
def display_widget():
    with flushed_registry():
        @register_binding_controller(
            klassname="Editor",
            binding_type=(StringBinding, FloatBinding))
        class DisplayWidget(BaseBindingController):
            model = Instance(UniqueModel)

            @staticmethod
            def initialize_model(proxy, model):
                model.attributes = proxy.binding.attributes

            def create_widget(self, parent=None):
                return QWidget(parent)

        yield DisplayWidget


def test_scene_dnd_handler(gui_app, display_widget, mocker):
    """Test the scene drag and drop handler"""
    schema = TypeObject.getClassSchema()
    proxy = get_property_proxy(schema, "prop")
    state_proxy = get_property_proxy(schema, "state")
    assert proxy.binding is not None
    assert state_proxy.binding is not None
    scene_model = SceneModel(children=[])
    scene_view = SceneView(model=scene_model)

    # 1. Reconfigurable item
    items = dragged_configurator_items([proxy])
    assert items is not None

    event = QDropEvent(QPoint(0, 0), Qt.CopyAction, items,
                       Qt.LeftButton, Qt.NoModifier, QEvent.Drop)
    handler = ConfigurationDropHandler()
    assert handler.can_handle(event)
    assert len(scene_model.children) == 0

    def get_proxy(*args, **kwargs):
        return proxy

    mocker.patch(GET_PROXY_PATH, new=get_proxy)
    handler.handle(scene_view, event)

    assert len(scene_model.children) == 1
    assert isinstance(scene_model.children[0], BoxLayoutModel)
    model = scene_model.children[0].children[1]
    assert isinstance(model, UniqueModel)
    assert model.attributes["minInc"] == -1000
    assert model.attributes["maxInc"] == 1000

    # 2. Readonly param
    items = dragged_configurator_items([state_proxy])
    assert items is not None

    event = QDropEvent(QPoint(0, 0), Qt.CopyAction, items,
                       Qt.LeftButton, Qt.NoModifier, QEvent.Drop)
    assert handler.can_handle(event)
    assert len(scene_model.children) == 1

    def get_proxy(*args, **kwargs):
        return state_proxy

    mocker.patch(GET_PROXY_PATH, new=get_proxy)
    handler.handle(scene_view, event)

    assert len(scene_model.children) == 2
    assert isinstance(scene_model.children[1], BoxLayoutModel)
    model = scene_model.children[1].children[1]
    assert isinstance(model, UniqueModel)
    assert model.attributes["displayType"] == "State"

    scene_view.destroy()



# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
from pytestqt.modeltest import ModelTester
from qtpy.QtCore import QItemSelectionModel, QPoint, Qt

from karabogui.dialogs.api import DevelopmentTopologyDialog
from karabogui.testing import development_system_hash, singletons
from karabogui.topology.system_topology import SystemTopology


def test_development_topology_dialog(gui_app, mocker):
    topology = SystemTopology()
    topology.initialize(development_system_hash())
    manager = mocker.Mock()
    with singletons(topology=topology, manager=manager):
        dialog = DevelopmentTopologyDialog()
        assert not dialog.isModal()
        model = dialog.tree_view.model()
        assert model is not None
        # We have 1 client
        assert model.rowCount() == 1
        client_index = model.index(0, 0)
        assert client_index.data() == "devserver"

        assert model.headerData(
            0, Qt.Horizontal, Qt.DisplayRole) == "Server Id"
        assert model.headerData(1, Qt.Horizontal, Qt.DisplayRole) == "Host"

        flag = model.flags(client_index)
        assert flag & Qt.ItemIsEnabled == Qt.ItemIsEnabled
        assert flag & Qt.ItemIsSelectable == Qt.ItemIsSelectable
        assert flag & Qt.ItemIsEditable != Qt.ItemIsEditable

        # Check context menu without selection
        pos = QPoint(0, 0)
        menu = mocker.patch("karabogui.dialogs.client_topology.QMenu")
        dialog._context_menu(pos)
        menu.assert_not_called()

        # Select the client and kill!
        selection = dialog.tree_view.selectionModel()
        selection.setCurrentIndex(client_index,
                                  QItemSelectionModel.ClearAndSelect)
        dialog.onKillInstance()
        assert manager.shutdownServer.call_count == 1
        manager.shutdownServer.assert_called_with("devserver", parent=dialog)


def test_model_tester():
    topology = SystemTopology()
    topology.initialize(development_system_hash())
    with singletons(topology=topology):
        dialog = DevelopmentTopologyDialog()
        model = dialog.tree_view.model()
        tester = ModelTester(None)
        tester.check(model)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Undefined

from karabo.native import Configurable, Float, Int32
from karabogui.binding.api import (
    DeviceClassProxy, PropertyProxy, ProxyStatus, build_binding)
from karabogui.indicators import (
    ALL_OK_COLOR, PROPERTY_ALARM_COLOR, PROPERTY_WARN_COLOR)
from karabogui.testing import set_proxy_value

from ..label import DisplayAlarmFloat, DisplayAlarmInteger


class Object(Configurable):
    number = Float(
        displayType="format|fmt=f&decimals=4")
    integer = Int32()


def test_alarm_numbers(gui_app):
    schema = Object.getClassSchema()
    binding = build_binding(schema)
    device = DeviceClassProxy(binding=binding, server_id="Fake",
                              status=ProxyStatus.OFFLINE)
    proxy = PropertyProxy(root_proxy=device, path="number")

    controller = DisplayAlarmFloat(proxy=proxy)
    # Initialize model for testing
    model = controller.model
    controller.initialize_model(proxy, model)
    controller.create(None)

    # Attributes not available anymore
    assert model.warnLow == Undefined
    assert model.warnHigh == Undefined
    assert model.alarmHigh == Undefined
    assert model.alarmLow == Undefined

    assert controller.fmt == "{:.4f}"
    model.warnLow = -1.0
    model.warnHigh = 1.0
    model.alarmHigh = 2.0
    model.alarmLow = -2.0

    set_proxy_value(proxy, "number", 0.75)
    assert controller.internal_widget.text() == "0.7500"
    assert controller.bg_color == ALL_OK_COLOR
    set_proxy_value(proxy, "number", 3.0)
    assert controller.bg_color == PROPERTY_ALARM_COLOR
    set_proxy_value(proxy, "number", 0.75)
    assert controller.bg_color == ALL_OK_COLOR
    set_proxy_value(proxy, "number", 1.2)
    assert controller.bg_color == PROPERTY_WARN_COLOR

    proxy = PropertyProxy(root_proxy=device, path="integer")
    controller = DisplayAlarmInteger(proxy=proxy)
    # Initialize model for testing
    model = controller.model
    controller.initialize_model(proxy, model)
    controller.create(None)

    assert model.warnLow == Undefined
    assert model.warnHigh == Undefined
    assert model.alarmHigh == Undefined
    assert model.alarmLow == Undefined

    model.warnLow = -1
    model.warnHigh = 1
    model.alarmHigh = 3
    model.alarmLow = -3

    set_proxy_value(proxy, "integer", 1)
    assert controller.internal_widget.text() == "1"
    assert controller.bg_color == ALL_OK_COLOR
    set_proxy_value(proxy, "integer", 4)
    assert controller.bg_color == PROPERTY_ALARM_COLOR
    set_proxy_value(proxy, "integer", 0)
    assert controller.bg_color == ALL_OK_COLOR
    set_proxy_value(proxy, "integer", 2)
    assert controller.bg_color == PROPERTY_WARN_COLOR



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np

from karabogui.graph.common.fitting import gaussian_fit
from karabogui.graph.common.formatting import (
    table_body, table_header, table_row)


def test_formatting():
    assert table_header() == "<th align='left' width='50'>&nbsp;</th>"
    assert table_body() == "<th align='left' width='50'>&nbsp;</th>"
    assert table_row() == "<tr></tr>"

    expected = "<th align='left' width='50'>Gaussian Fit Result</th>"
    assert table_header(header="Gaussian Fit Result") == expected

    expected = ("<th align='left' width='50'>Gaussian Fit Result</th><th "
                "align='left' width='50'>one</th><th align='left' "
                "width='50'>two</th>")
    assert table_header(
        header="Gaussian Fit Result", tabs=("one", "two")) == expected

    expected = ("<th align='left' width='50'>one</th><td "
                "width='50'>t</td><td width='50'>w</td><td width='50'>o</td>")
    assert table_body("one", "two") == expected


def test_gaussian_fitting():
    x = np.array([1, 2, 3, 4, 5])
    h = 1.0
    x0 = 1.0
    sigma = 1.0
    offset = 1.0
    params = gaussian_fit(x, h, x0, sigma, offset)
    expected = np.array([2, 1.60653066, 1.13533528, 1.011109, 1.00033546])
    params = np.round(params, 5)
    expected = np.round(expected, 5)
    assert np.array_equal(params, expected)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from traits.api import (
    ArrayOrNone, HasStrictTraits, Int, Property, on_trait_change)

DEFAULT_STACK = 100


class RollImage(HasStrictTraits):
    """Class to construct a waterfall image from an array

    This class will build a rolling image from the input arrays up to a limit
    of `stack` size. The default stack size is `100`.

    Arrays can be added via:

    image = RollImage()
    image.add(...)

    The internal data is automatically reset if the `stack` size changes or
    the array input changes in `shape` or `dtype`.

    The rolling image can then be derived via the `data` property:

        data = image.data
    """

    data = Property
    stack = Int

    # Internal traits
    _slice_counter = Int
    _data = ArrayOrNone

    def _stack_default(self):
        return DEFAULT_STACK

    def _get_data(self):
        if self._data is None:
            return None

        return self._data[0:self._slice_counter + 1]

    @on_trait_change('stack')
    def invalidate(self):
        self._data = None

    # Public interface
    # -----------------------------------------------------------------------

    def reset(self):
        self.invalidate()

    def add(self, value):
        if value.size == 0:
            self._data = None
            return

        finite = np.isfinite(value)
        if not np.all(finite):
            value = np.array(value, copy=True)
            value[~finite] = 0

        needs_reset = (self._data is None
                       or (value.shape[0] != self._data.shape[1])
                       or (value.dtype != self._data.dtype))
        if needs_reset:
            self._data = np.vstack([value for _ in range(self.stack)])
            self._slice_counter = 0

        elif self._slice_counter < self.stack:
            self._slice_counter += 1
        self._data = np.roll(self._data, 1, axis=0)
        self._data[0] = value



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

from qtpy.QtGui import QFont, QFontMetrics

from karabo.common.scenemodel.api import SCENE_FONT_FAMILY, SCENE_FONT_SIZE
from karabogui.const import GUI_DPI_FACTOR


def get_qfont(font_string='', adjust_size=True):
    assert type(font_string) is str

    q_font = QFont()
    if font_string:
        q_font.fromString(font_string)
    else:
        # Enforce scene font size. We do not depend on the QApplication font
        # anymore as it scales depending on the OS
        q_font.setPointSize(SCENE_FONT_SIZE)

    # Correct font size for Mac OSX
    if adjust_size:
        font_size = get_font_size_from_dpi(q_font.pointSize())
        q_font.setPointSize(font_size)

    # Compare style hint and replace accordingly
    if q_font.family() not in FONT_FAMILIES:
        font = FONT_STYLE_HINTS.get(q_font.styleHint(), SCENE_FONT_FAMILY)
        q_font.setFamily(font)

    return q_font


def get_font_metrics():
    """Return the FontMetric for the default QFont"""
    q_font = get_qfont()
    return QFontMetrics(q_font)


def substitute_font(model):
    if "font" not in model.trait_names():
        return

    q_font = get_qfont(model.font, adjust_size=False)
    model.trait_setq(font=q_font.toString())


def _get_font_filenames():
    # Returns the list of font filenames relative to the fonts/ folder.
    # This is then used when adding the fonts in the font database."""
    folder = Path(__file__).parent
    return [str(filename) for filename in folder.rglob("*.ttf")]


FONT_FILENAMES = _get_font_filenames()

# This maps the font family to a Qt style hint to ease font replacement
FONT_STYLE_HINTS = {
    QFont.Helvetica: "Source Sans Pro",
    QFont.Times: "Source Serif Pro",
    QFont.Courier: "Source Code Pro"
}

FONT_FAMILIES = sorted(list(set(FONT_STYLE_HINTS.values())))

# This maps the font family to a Qt style hint to ease font replacement
FONT_STYLE_ALIAS = {
    QFont.Helvetica: "Sans Serif",
    QFont.Times: "Serif",
    QFont.Courier: "Monospaced"
}

FONT_ALIAS = sorted(list(set(FONT_STYLE_ALIAS.values())))


def get_font_from_alias(alias):
    for hint, name in FONT_STYLE_ALIAS.items():
        if alias == name:
            return FONT_STYLE_HINTS[hint]

    return SCENE_FONT_FAMILY


def get_alias_from_font(font_family):
    for hint, family in FONT_STYLE_HINTS.items():
        if font_family == family:
            return FONT_STYLE_ALIAS[hint]

    return SCENE_FONT_FAMILY


def get_font_size_from_dpi(size):
    """Calculates the effective font size from the OS DPI.
    As Qt and the majority of the OSes use DPI=96, we should add some
    scaling factor on the sizes for Mac DPI, as it has DPI=72.
    We round off to the nearest integer."""
    return round(size * GUI_DPI_FACTOR)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pyqtgraph import GraphicsLayoutWidget, LabelItem
from traits.api import Dict, Enum, HasStrictTraits, Instance

from karabogui.graph.common.api import AuxPlots, ImageRegion

from .histogram.controller import HistogramAggregator
from .profiling.controller import ProfileAggregator

# Map of the enum and the aux plot aggregator
AGGREGATOR_MAP = {
    AuxPlots.ProfilePlot: ProfileAggregator,
    AuxPlots.Histogram: HistogramAggregator
}

# Grid orientation of the image layout, where the center is (1, 1)
GRID_ORIENTATION = {
    "top": (0, 1),
    "left": (1, 0),
}


class AuxPlotsController(HasStrictTraits):
    """This is used by the image base to control the AuxPlots as one entity.
       This contains the aggregators of each added type, which, in turn,
       contains the controllers of each plot contents. This also has the
       shared labelItem that shows the statistics in an HTML format."""

    # The image layout that contains the image plot and will house the
    # aux plot items
    image_layout = Instance(GraphicsLayoutWidget)

    # The region from the image plot
    _region = Instance(ImageRegion)

    # The current aux plot. Default value is `AuxPlots.NoPlot`
    current_plot = Enum(*AuxPlots)

    # The added aggregators: e.g., `{AuxPlots.Histogram: HistogramAuxPlot()}`
    _aggregators = Dict

    # The UI item that shows HTML tables of the statistics
    labelItem = Instance(LabelItem, kw={"justify": "left"})

    # -----------------------------------------------------------------------
    # Public methods

    def add(self, plot_type, **config):
        """Set the aux plots of specified plot type to the image layout"""
        # Add label item for aux plots stats. This will be shown in the topleft
        if plot_type in self._aggregators:
            return

        # Remove set plots if any
        aggregator_klass = AGGREGATOR_MAP[plot_type]
        aggregator = aggregator_klass(config=config)
        aggregator.on_trait_change(self._show_stats, 'stats')

        self._aggregators[plot_type] = aggregator
        return aggregator

    def destroy(self):
        """Delete all the items and the controller gracefully"""
        # Check if already cleared
        if not len(self.image_layout.ci.items):
            return

        self.image_layout.clear()
        for aggregator in self._aggregators.values():
            aggregator.destroy()

        self.labelItem.deleteLater()

    def process(self, region):
        self._region = region
        if self.current_plot is not AuxPlots.NoPlot:
            aux_plot = self._aggregators.get(self.current_plot)
            aux_plot.process(region)

    def set_axes(self, x_data, y_data):
        for aux_plot in self._aggregators.values():
            aux_plot.set_axes(x_data, y_data)

    def set_config(self, plot=None, **config):
        aggregator = self._aggregators.get(plot)
        if aggregator is not None:
            aggregator.trait_set(**config)

    # -----------------------------------------------------------------------
    # Private methods

    def _add_item(self, plot_item, orientation="top"):
        row, col = GRID_ORIENTATION[orientation]
        self.image_layout.addItem(plot_item, row, col)
        if orientation == "top":
            self.image_layout.ci.layout.setRowStretchFactor(row, 1)
            self.image_layout.ci.layout.setRowMinimumHeight(row, 100)
        elif orientation == "left":
            self.image_layout.ci.layout.setColumnStretchFactor(col, 1)
            self.image_layout.ci.layout.setColumnMinimumWidth(col, 100)

    def _remove_item(self, item, orientation=None):
        if item in self.image_layout.ci.items.keys():
            self.image_layout.removeItem(item)
        # Collapse the grid space by removing the stretch factor of the
        # emptied space
        row, col = GRID_ORIENTATION[orientation]
        if orientation == "top":
            self.image_layout.ci.layout.setRowStretchFactor(row, 0)
            self.image_layout.ci.layout.setRowMinimumHeight(row, 0)
        elif orientation == "left":
            self.image_layout.ci.layout.setColumnStretchFactor(col, 0)
            self.image_layout.ci.layout.setColumnMinimumWidth(col, 0)

    def _show_stats(self, stats):
        """Add the labelItem only if there's stats shown"""
        label_added = self.labelItem in self.image_layout.ci.items.keys()
        if stats is not None:
            if not label_added:
                # Span the label on unoccupied orientation.
                # The item spans by 1 row and column by default,
                # but now we span to an unoccupied space if available.
                # For instance, with histogram aux plot, that does not have a
                # left plot, we span the label by two columns.
                aggregator = self._aggregators[self.current_plot]
                plot_orientations = aggregator.controllers.keys()
                row_span = 1 + ("left" not in plot_orientations)
                col_span = 1 + ("top" not in plot_orientations)
                self.image_layout.addItem(self.labelItem, 0, 0,
                                          row_span, col_span)
            self.labelItem.item.setHtml(stats.html)
        elif label_added:
            self.labelItem.setText('')
            self.image_layout.removeItem(self.labelItem)
        self.labelItem.resizeEvent(None)

    # -----------------------------------------------------------------------
    # Trait handlers

    def _current_plot_changed(self, old, new):
        # Check if requested plot is in instantiated aggregators
        if new not in self._aggregators:
            new = AuxPlots.NoPlot

        # Remove previous plot items
        if old != AuxPlots.NoPlot:
            aggregator = self._aggregators[old]
            for orientation, plot in aggregator.controllers.items():
                self._remove_item(plot.plotItem, orientation)
            self._show_stats(None)

        if new != AuxPlots.NoPlot:
            # Add new plot items
            aggregator = self._aggregators[new]
            for orientation, plot in aggregator.controllers.items():
                self._add_item(plot.plotItem, orientation)
            self._show_stats(None)

        # Process the changes
        if self._region is not None:
            self.process(self._region)
